/**
  * File: BehaviourOnTransferImpl.java
  * Content: Abstract class over the implementation classes implementing the behaviour models (on transfer state)
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.domains.*;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;

import sensei.util.Error;
import sensei.util.IntWrapper;

import java.util.Iterator;
import java.util.Map;

/**
 * Abstract class over the implementation classes implementing the behaviour models (on transfer state)
 */
public abstract class BehaviourOnTransferImpl implements EventsQueueUser, StateSenderThreadUser
{

  //*************************************************************************************
  //**************************** CONSTRUCTOR ********************************************
  //*************************************************************************************

  public BehaviourOnTransferImpl(GroupHandler groupHandler,
                                 DomainGroupHandler domainGroupHandler,
                                 DomainsUser user,
                                 SubgroupsManager subgroupsManager,
                                 PropertyManager propertyManager,
                                 DomainMessagesHandler messageHandler,
                                 int memberIdentity,
                                 View initialView,
                                 BehaviourOnViewChanges behaviour)
  {
    Trace.code("BehaviourOnTransferImpl.java -> BehaviourOnTransferImpl ( GroupHandler groupHandler , DomainGroupHandler domainGroupHandler , DomainsUser user , SubgroupsManager subgroupsManager , PropertyManager propertyManager , DomainMessagesHandler messageHandler , int memberIdentity , View initialView , BehaviourOnViewChanges behaviour )");
    this.groupHandler = groupHandler;
    this.subgroupsManager = subgroupsManager;
    this.domainGroupHandler = domainGroupHandler;
    this.domainsUser = user;
    this.propertyManager = propertyManager;
    this.messageHandler = messageHandler;
    this.behaviour=behaviour;
    expulsionReason = DomainExpulsionReason.GMSLeaveEvent;
    coordination=null;
    memberId = new Integer(memberIdentity);
    membersStateHandler=new MembersStateHandler(initialView.viewId, initialView.members, initialView.newMembers, user);
    statelessMemberInfo=membersStateHandler.getStatelessInfo(memberId);
    messenger = new DomainsFactory(groupHandler, behaviour, user.supportsCoordinatorElection(),
      propertyManager!=null, user.supportsDynamicBehaviour());
    dynamicSubgroupsHandler = new DynamicSubgroupsHandler(memberIdentity, user, subgroupsManager, messenger);

    viewId=initialView.viewId;
    coordinations=null;
    stateSenderThread=null;
    eventsQueue = new EventsQueue(this);
    onBlockedView = false;

    if (propertyManager==null)
    {
      nullMemberProperties = new MemberProperties [0];
      nullProperties = new Property[0];
    }
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public void freeResources()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void freeResources ( )");
    if (memberId!=null)
    {
      memberId=null;
      eventsQueue.freeResources();
      eventsQueue=null;
      propertyManager=null;
      subgroupsManager=null;
      membersStateHandler.freeResources();
      membersStateHandler=null;
      behaviour=null;
      dynamicSubgroupsHandler.freeResources();
      dynamicSubgroupsHandler=null;
      coordination=null;
      messageHandler=null;
      messenger.freeResources();
      messenger=null;
      groupHandler=null;
      domainGroupHandler=null;
      domainsUser=null;
      if (stateSenderThread!=null)
      {
        stateSenderThread.freeResources();
        stateSenderThread=null;
      }
      if (coordinations!=null)
      {
        coordinations.freeResources();
        coordinations=null;
      }
      nullMemberProperties=null;
      nullProperties=null;
      expulsionReason=null;
      if (stateMemberInfo!=null)
      {
        stateMemberInfo.freeResources();
        stateMemberInfo=null;
      }
      statelessMemberInfo=null;
    }
  }

  //*************************************************************************************
  //**************************** GET STATEFUL MEMBERS ***********************************
  //*************************************************************************************

  /**
   * Returns the list of members with state, raising an exception if the member has no state
   */
  public int[] getStatefulMembers() throws MemberStateException
  {
    Trace.code("BehaviourOnTransferImpl.java -> int [ ] getStatefulMembers ( )");
    if (!hasState())
      throw new MemberStateException(MemberStateExceptionReason.MemberWithoutState);
    return membersStateHandler.getMembers();
  }

  //*************************************************************************************
  //**************************** GET EVENTS QUEUE ***************************************
  //*************************************************************************************

  /**
   * The events queue is created at this level, but must be accesible from the DomainGroupHandler
   */
  public EventsQueue getEventsQueue()
  {
    Trace.code("BehaviourOnTransferImpl.java -> EventsQueue getEventsQueue ( )");
    return eventsQueue;
  }

  //*************************************************************************************
  //**************************** GET DYNAMIC SUBGROUPS HANDLER **************************
  //*************************************************************************************

  /**
   * The dynamic subgroups handler is created at this level, but must be accesible from the DomainGroupHandler
   */
  public DynamicSubgroupsHandler getDynamicSubgroupsHandler()
  {
    Trace.code("BehaviourOnTransferImpl.java -> DynamicSubgroupsHandler getDynamicSubgroupsHandler ( )");
    return dynamicSubgroupsHandler;
  }

  //*************************************************************************************
  //**************************** EVENTS QUEUE USER OPERATIONS ***************************
  //*************************************************************************************

  /**
   * Event thrown on first place, to inform on the group acceptance
   */
  public void memberAccepted(View initialView)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void memberAccepted ( View initialView )");
    Trace.memberAccepted();

    try
    {
      //informs the subgroups about the member acceptance
      subgroupsManager.memberAccepted(memberId.intValue(), domainGroupHandler, initialView);
      //informs the domain about the member acceptance
      domainsUser.domainAccepted(memberId.intValue());
      //check if the member can directly achieve state
      checkStateAssumption();
    }
    catch(SubgroupException ex)
    {
      subgroupError(ex);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  /**
   * Event received from the EventsQueueUser.
   */
  public void processViewEvent(View view)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processViewEvent ( View view )");

    //The reaction is the same in all the cases, the view is sent to the subgroups
    onBlockedView=false;
    viewId=view.viewId;
    try
    {
      subgroupsManager.installView(view);
    }
    catch(SubgroupException ex)
    {
      subgroupError(ex);
    }
  }

  /**
   * Event thrown after receiving an expuslsion event
   */
  public void processExpulsionEvent()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processExpulsionEvent ( )");
    //The reaction is the same in all the cases, inform the subgroups
    try
    {
      subgroupsManager.memberExpulsed();
      domainsUser.domainExpulsed(expulsionReason);
    }
    catch(SubgroupException ex)
    {
      subgroupError(ex);
    }
  }

  /**
   * Event thrown to process the given PTP message
   */
  public void processPTPMessageEvent(int sender, Message message)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processPTPMessageEvent ( int sender , Message message )");
    //if message is not state message, send it to the app
    if (message instanceof MessageS)
      processMessageS(sender, (MessageS) message);
    else if (message instanceof MessageC)
      processMessageC(sender, (MessageC) message);
    else if (message instanceof MessageP)
      processMessageP(sender, (MessageP) message);
    else
      messageHandler.processPTPMessageEvent(sender, message);
  }

  /**
   * Event thrown to process the given cast message
   */
  public void processCastMessageEvent(int sender, Message message)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processCastMessageEvent ( int sender , Message message )");
    try
    {
      if (message instanceof MessageE)
        processMessageE(sender, (MessageE) message);
      else if (message instanceof MessageMemberProperties)
        processMessageMemberProperties(sender, (MessageMemberProperties) message);
      else if (message instanceof MessageRemoveSubgroup)
        dynamicSubgroupsHandler.processMessageRemoveSubgroup(sender, (MessageRemoveSubgroup) message);
      else if (message instanceof MessageCreateSubgroup)
      {
        if (!dynamicSubgroupsHandler.processMessageCreateSubgroup(sender, (MessageCreateSubgroup) message))
        {
          expulsionReason = DomainExpulsionReason.WrongSubgroupsTypes;
          leave();
        }
      }
      else
        messageHandler.processCastMessageEvent(sender, message);
    }
    catch(SubgroupException sex)
    {
      subgroupError(sex);
    }
  }

  abstract public boolean hasStrongSTMessageConsistency();
  abstract void processMessageC(int target, MessageC message);
  abstract void processMessageE(int target, MessageE message);
  abstract void processMessageP(int target, MessageP message);
  abstract void processMessageS(int target, MessageS message);
  abstract public void startTransaction(Monitor monitor)
    throws MonitorException, TransactionException, MemberStateException;
  abstract public void endTransaction()
      throws MonitorException, TransactionException, MemberStateException;


  //*************************************************************************************
  //**************************** PROCESS MESSAGE MEMBER PROPERTIES **********************
  //*************************************************************************************

  void processMessageMemberProperties(int sender, MessageMemberProperties message)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processMessageMemberProperties ( int sender , MessageMemberProperties message )");

    if (propertyManager==null)
      Trace.errorReceivingMessageMemberProperties("Group supports no properties");
    else if (propertyManager.setProperties(sender, message.props, false))
    {
      Trace.receivedMessageMemberProperties(sender, message);
      reportPropertiesUpdate(sender);
    }
    else
      Trace.errorReceivingMessageMemberProperties("member is not known "+sender);
  }

  void reportPropertiesUpdate(int sender)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void reportPropertiesUpdate ( int sender )");
    domainsUser.propertiesUpdated(sender);
  }

  //*************************************************************************************
  //**************************** CHECK GROUP FEATURES ***********************************
  //*************************************************************************************

  /**
   * Check that the group features matches. If not, it does a leave automatically
   */
  boolean checkGroupFeatures(MessageP message)
  {
    Trace.code("BehaviourOnTransferImpl.java -> boolean checkGroupFeatures ( MessageP message )");

    try
    {
      //if any member registers a coordinator, every member must do it
      if (message.coordinatorElectorRegistered != domainsUser.supportsCoordinatorElection())
        expulsionReason = DomainExpulsionReason.WrongCoordinatorElectionPolicy;
      else if (message.behaviour!=behaviour)
        expulsionReason = DomainExpulsionReason.WrongBehaviourMode;
      else if ((message.props.length==0)!=(propertyManager==null))
        expulsionReason = DomainExpulsionReason.WrongPropertyAllowance;
      else if (message.dynamicSubgroupsAllowed!=domainsUser.supportsDynamicBehaviour())
        expulsionReason = DomainExpulsionReason.WrongDynamicPolicy;
      else if (!subgroupsManager.checkStaticSubgroups(message.subgroupsInformation))
        expulsionReason = DomainExpulsionReason.WrongStaticSubgroupsComposition;
      else if (message.dynamicSubgroupsAllowed)
      {
        if (!dynamicSubgroupsHandler.createDynamicSubgroups(message.nextDynamicSubgroupId, message.subgroupsInformation))
            expulsionReason = DomainExpulsionReason.WrongSubgroupsTypes;
      }

      Trace.checkedGroupFeatures(expulsionReason);

      if (expulsionReason!=DomainExpulsionReason.GMSLeaveEvent)
        leave();
    }
    catch(SubgroupException sex)
    {
      subgroupError(sex);
    }
    return expulsionReason==DomainExpulsionReason.GMSLeaveEvent;
  }

  void leave()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void leave ( )");
    try
    {
      groupHandler.leaveGroup();
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  //*************************************************************************************
  //**************************** BECOMING STATEFULL RELATED OPERATIONS ******************
  //*************************************************************************************

  /**
   * To be called on stateless members, it is checked if the member can assume state
   */
  void checkStateAssumption()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void checkStateAssumption ( )");
    //if there are no members with state and this is the oldest member, assume state
    Integer oldest = membersStateHandler.getOldestStatelessMember();
    if ((oldest!=null) && (oldest.equals(memberId)))
    {
      Trace.assumingState();
      //assume state
      assumeState();
      //check transfers to do
      checkTransferToDo();
    }
    else
      Trace.assumingState(oldest);
  }

  /**
   * Assumes the state
   */
  void assumeState()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void assumeState ( )");
    assert !hasState();

    membersStateHandler.assumeState(memberId);
    try
    {
      subgroupsManager.assumeState();
      if (propertyManager!=null)
        propertyManager.setGroupProperties();
      becameStatefull(true);
    }
    catch(SubgroupException ex)
    {
      subgroupError(ex);
    }
  }

  /**
   * Set of operartions to be performed when the member becomes statefull
   */
  void becameStatefull(boolean stateAssumed)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void becameStatefull ( boolean stateAssumed )");
    removeCoordinator(false);
    statelessMemberInfo=null;
    stateMemberInfo=membersStateHandler.getStateInfo(memberId);
    try
    {
      if (propertyManager!=null)
        propertyManager.setMessenger(messenger);
      domainsUser.stateObtained(stateAssumed);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  //*************************************************************************************
  //**************************** COORDINATION OPERATIONS ********************************
  //*************************************************************************************

  /**
   * Sets the coordinating member; it can be set to null directly
   */
  void setCoordinatingMember(Integer member)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void setCoordinatingMember ( Integer member )");
    assert hasState();
    coordination = member;
  }

  /**
   * Sets the coordinator; the replica info given by the coordinator must be specified. For this reason,
   * to specify a null coordinator, use removeCoordinator, as the information on the another members must be cleared
   */
  void setCoordinator(Integer member, ReplicaInformation[] replicasInfo)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void setCoordinator ( Integer member , ReplicaInformation [ ] replicasInfo )");
    assert !hasState();
    assert member!=null;
    coordination = member;
    membersStateHandler.rebuildInformation(replicasInfo);
  }

  /**
   * Removes the coordinator. If the coordinator is down, set the argument to true, to clear any member information
   * given by that member
   */
  void removeCoordinator(boolean coordinatorIsDown)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void removeCoordinator ( boolean coordinatorIsDown )");
    assert !hasState();
    coordination = null;
    if (coordinatorIsDown)
      membersStateHandler.clearInformation();
  }

  Integer getCoordinatingMember()
  {
    Trace.code("BehaviourOnTransferImpl.java -> Integer getCoordinatingMember ( )");
    assert hasState();
    return coordination;
  }

  Integer getCoordinator()
  {
    Trace.code("BehaviourOnTransferImpl.java -> Integer getCoordinator ( )");
    assert !hasState();
    return coordination;
  }

  /**
   * This method looks for the next member to be coordinated, not the one set by the class
   */
  Integer getMemberToCoordinate()
  {
    Trace.code("BehaviourOnTransferImpl.java -> Integer getMemberToCoordinate ( )");
    assert hasState();
    return getMemberInfo().getMemberToCoordinate();
  }

  //*************************************************************************************
  //**************************** HANDLING COORDINATIONS *********************************
  //*************************************************************************************

  /**
   * Initialize the coordinations map, which is automatically stored; it is also returned
   * This method has only sense on stateless members
   */
  CoordinationsMap initCoordinationsMap() throws SubgroupException
  {
    Trace.code("BehaviourOnTransferImpl.java -> CoordinationsMap initCoordinationsMap ( )");
    //can only be called on non state members
    assert !hasState();
    coordinations = subgroupsManager.initCoordinations();
    return coordinations;
  }

  /**
   * Checks the coordinations map against new/removed subgroups, which is automatically stored; it is also returned
   * This method has only sense on stateless members
   */
  CoordinationsMap checkCoordinationsMap() throws SubgroupException
  {
    Trace.code("BehaviourOnTransferImpl.java -> CoordinationsMap checkCoordinationsMap ( )");
    //can only be called on non state members
    assert !hasState();
    coordinations = subgroupsManager.checkDynamicCoordinations(coordinations);
    return coordinations;
  }

  /**
   * This method returns all the coordinations, including those were anyValidCoordination is false
   * This method has only sense on stateless members
   */
  SubgroupCoordination[] getSubgroupCoordinations()
  {
    Trace.code("BehaviourOnTransferImpl.java -> SubgroupCoordination [ ] getSubgroupCoordinations ( )");
    //can only be called on non state members
    assert !hasState();
    assert coordinations!=null;
    int size=coordinations.coordinationsMap.size();
    SubgroupCoordination[] ret = new SubgroupCoordination[size];
    Iterator it = coordinations.coordinationsMap.values().iterator();
    while(it.hasNext())
      ret[--size]=(SubgroupCoordination) (it.next());
    assert size==0;
    return ret;
  }

  /**
   * This method returns all the coordinations, including those were anyValidCoordination is false
   * This method has only sense on stateless members
   */
  CoordinationsMap getCoordinations(boolean initIfNotExisting, boolean checkSubgroupsChanges) throws SubgroupException
  {
    Trace.code("BehaviourOnTransferImpl.java -> CoordinationsMap getCoordinations ( boolean initIfNotExisting , boolean checkSubgroupsChanges )");
    assert !hasState();
    CoordinationsMap ret = coordinations;
    if (ret==null)
    {
      assert initIfNotExisting;
      if (initIfNotExisting)
        ret=initCoordinationsMap();
    }
    else if (checkSubgroupsChanges)
      ret=checkCoordinationsMap();
    return ret;
  }

  /**
   * Set the coordinations given by the server on the message S.
   * It fills the map subgroupStates with the states included as well
   * It does nothing and returns null if the message is considered to be invalid
   */
  CoordinationsMap setCoordinations(SubgroupState[] serverStates, Map subgroupStates)
  {
    Trace.code("BehaviourOnTransferImpl.java -> CoordinationsMap setCoordinations ( SubgroupState [ ] serverStates , Map subgroupStates )");
    assert !hasState();
    CoordinationsMap ret=null;
    int size=serverStates.length;
    if ((coordinations!=null) && (coordinations.coordinationsMap.size()>=size))
    {
      coordinations.reset();
      subgroupStates.clear();
      for(int i=0;i<size;i++)
      {
        SubgroupState state = serverStates[i];
        IntWrapper subgroup = new IntWrapper(state.id);
        coordinations.coordinationsMap.put(subgroup, new SubgroupCoordination(subgroup.i, state.phase));
        subgroupStates.put(subgroup, state.subState);
      }
      ret=coordinations;
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** CHECK TRANSFERS TO DO **********************************
  //*************************************************************************************

  /**
   * Verifies the transfers to be done
   */
  void checkTransferToDo()
  {
    Trace.code("BehaviourOnTransferImpl.java -> void checkTransferToDo ( )");
    assert getCoordinatingMember()==null;
    //check if there are members to coordinate
    Integer coordinating = getMemberToCoordinate();
    Trace.memberToCoordinate(coordinating);

    if (coordinating==null)
    {
      //if no members to coordinate, the events queue should be unblocked. The events to unblock are those
      //from this view
      eventsQueue.unblock(false, true);
    }
    else
    {
      setCoordinatingMember(coordinating);
      if (!sendMessageP()) //The message must been sent on the same view!!
        setCoordinatingMember(null);
    }
  }

  //*************************************************************************************
  //**************************** PROCESS VIEW CHANGES ***********************************
  //*************************************************************************************

  /**
   * Modifies the group knowledge after a view change
   */
  void processViewChanges(View view)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void processViewChanges ( View view )");
    membersStateHandler.removeMembers(view.expulsedMembers);
    membersStateHandler.addMembers(view.newMembers, view.viewId);
  }

  void setViewAsBlocked(){onBlockedView=true;}

  boolean isViewAlreadyBlocked(){return onBlockedView;}

  //*************************************************************************************
  //**************************** SYNC TRANSFER ******************************************
  //*************************************************************************************

  /**
   * Requests to the subgoups for the synchronization and sends the corresponding message C
   * @param sendProperties is true if the properties must be sent. If the member is not using properties,
   *    this parameter is discarded
   */
  void syncTransfer(boolean sendProperties)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void syncTransfer ( boolean sendProperties )");
    assert !hasState();
    int target=getCoordinator().intValue();
    try
    {
      //inform to the subgrups that the transfer start. The coordinations map is directly modified
      //if sendProperties would be true, this could be the first transfer, and it can be necessary to init the coord.
      //Finally, subgroups are checked against dynamic changes
      subgroupsManager.syncTransfer(target, getCoordinations(sendProperties, true));

      Trace.sendMessageC(target);

      Property [] properties;
      boolean changed;
      if (propertyManager==null)
      {
        changed=false;
        properties = nullProperties;
      }
      else
      {
        PropertiesAndChange props = propertyManager.getOwnProperties(sendProperties);
        properties = props.properties;
        changed = props.changed;
      }
      messenger.sendMessageC(target, viewId, getSubgroupCoordinations(), changed, properties);
    }
    catch(SubgroupException sex)
    {
      subgroupError(sex);
    }
    catch(Exception ex)
    {
      Trace.handledException(ex);
    }
  }

  //*************************************************************************************
  //**************************** SEND MESSAGE P *****************************************
  //*************************************************************************************

  /**
   * Sends the message P, returning true if the message is sent on the current view
   */
  boolean sendMessageP()
  {
    Trace.code("BehaviourOnTransferImpl.java -> boolean sendMessageP ( )");
    boolean ret;
    int target=getCoordinatingMember().intValue();

    MemberProperties[] groupProperties;
    if (propertyManager==null)
      groupProperties=nullMemberProperties;
    else
      groupProperties = propertyManager.getGroupProperties();

    Trace.sendMessageP(target);

    try
    {
      ret=messenger.sendMessageP(target, viewId, dynamicSubgroupsHandler.getNextDynamicSubgroupId(),
        membersStateHandler.getReplicasInformation(), groupProperties, subgroupsManager.getSubgroupProperties());
    }
    catch(Exception ex)
    {
      Trace.handledException(ex);
      ret=false;
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** STATE SENDER THREAD METHODS ****************************
  //*************************************************************************************

  /**
   * The thread is started knowing the message C received from the client
   * doNotStartIfBlocked is true if the transfer must start only if the view is not blocked
   */
  synchronized void startStateSenderThread(MessageC message, boolean doNotStartIfBlocked)
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized void startStateSenderThread ( MessageC message , boolean doNotStartIfBlocked )");
    assert stateSenderThread==null;

    Property[] properties = (propertyManager==null)? nullProperties : message.props;

    stateSenderThread = new StateSenderThread(getCoordinatingMember().intValue(), message.coordinations,
        subgroupsManager, properties, messenger, this, viewId, doNotStartIfBlocked && onBlockedView);
  }

  /**
   * Pauses the stateSenderThread if running
   */
  synchronized void pauseStateSenderThread()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized void pauseStateSenderThread ( )");
    if (stateSenderThread!=null)
      stateSenderThread.pauseTransfer();
  }

  /**
   * Changes the view on the stateSenderThread.
   */
  synchronized void changeViewOnSenderThread()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized void changeViewOnSenderThread ( )");
    if (stateSenderThread!=null)
      stateSenderThread.changeViewId(viewId);
  }

  /**
   * Interrups the stateSenderThread if running
   */
  synchronized void interruptStateSenderThread()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized void interruptStateSenderThread ( )");
    if (stateSenderThread!=null)
      try
      {
        stateSenderThread.interruptTransfer();
      }
      catch(SubgroupException sex)
      {
        subgroupError(sex);
      }
  }

  /**
   * Continues the stateSenderThread if running; it also changes the view automatically
   */
  synchronized void continueStateSenderThread(MessageC message)
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized void continueStateSenderThread ( MessageC message )");

    if (stateSenderThread!=null)
      try
      {
        Property[] receivedProperties = message.containOwnProperties? message.props : null;
        stateSenderThread.changeViewId(viewId);
        stateSenderThread.continueTransfer(message.coordinations, receivedProperties);
      }
      catch(SubgroupException sex)
      {
        subgroupError(sex);
      }
  }

  /**
   * Stops the stateSenderThread if running, and returns the coordinatios currently sent to the client,
   * or null if there is not such a transfer
   */
  synchronized CoordinationsMap stopStateSenderThread()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized CoordinationsMap stopStateSenderThread ( )");
    CoordinationsMap ret=null;
    if (stateSenderThread!=null)
    {
      stateSenderThread.stopTransfer();
      ret=stateSenderThread.getCoordinations();
      stateSenderThread=null;
    }
    return ret;
  }

  /**
   * Completes the StateSenderThread, after having received a valid message E sent by the thread. It is
   * only needed when hasStrongSTMessageConsistency() returns true
   * returns the coordinatios currently sent to the client, or null if there is not such a transfer
   */
  synchronized CoordinationsMap completeStateSenderThread()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized CoordinationsMap completeStateSenderThread ( )");
    CoordinationsMap ret=null;
    if (stateSenderThread!=null)
    {
      stateSenderThread.completedTransfer();
      ret=stateSenderThread.getCoordinations();
      stateSenderThread=null;
    }
    return ret;
  }

  /**
   * Returns true if the transfer has started on the coordinator side. The coordinator is only really aware of
   * the coordination start after receiving the message C from the coordinated member, and that condition can be
   * checked using the method getStateSendCoordinations
   */
  synchronized boolean isTransferStarted()
  {
    Trace.code("BehaviourOnTransferImpl.java -> synchronized boolean isTransferStarted ( )");
    return stateSenderThread!=null;
  }

  /**
   * Called when there has been an error on the own subgroups. The transfer ends automatically
   */
  public void stateTransferThreadSubgroupError(SubgroupException sex)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void stateTransferThreadSubgroupError ( SubgroupException sex )");
    synchronized(this)
    {
      stateSenderThread=null;
    }
    subgroupError(sex);
  }

  //*************************************************************************************
  //**************************** ACCESS INFO OPERATIONS *********************************
  //*************************************************************************************

  MemberInfo getMemberInfo()
  {
    Trace.code("BehaviourOnTransferImpl.java -> MemberInfo getMemberInfo ( )");
    assert stateMemberInfo!=null;
    return stateMemberInfo;
  }

  StatelessMemberInfo getStatelessMemberInfo()
  {
    Trace.code("BehaviourOnTransferImpl.java -> StatelessMemberInfo getStatelessMemberInfo ( )");
    assert statelessMemberInfo!=null;
    return statelessMemberInfo;
  }

  boolean hasState(){return statelessMemberInfo==null;}

  //*************************************************************************************
  //**************************** SUBGROUP ERROR *****************************************
  //*************************************************************************************

  void subgroupError(SubgroupException sex)
  {
    Trace.code("BehaviourOnTransferImpl.java -> void subgroupError ( SubgroupException sex )");
    Trace.handledException(sex);
    //subgroup is informed
    try
    {
      domainsUser.offendingSubgroup(sex.subgroupId, sex.toString());
      expulsionReason = DomainExpulsionReason.SubgroupError;
      groupHandler.leaveGroup();
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  //identity of the member
  Integer memberId;

  //The current viewId is stored
  int viewId;

  //the events queue is created at this level, even when this class does not make much
  //(if any) direct use of such an object
  EventsQueue eventsQueue;

  //handler of the properties
  PropertyManager propertyManager; //null if properties are not allowed

  //subgroupsManager is used to make calls on the subgroups
  SubgroupsManager subgroupsManager;

  //the state of each member is handled through the MembersStateHandler instance
  MembersStateHandler membersStateHandler;

  //behaviour
  BehaviourOnViewChanges behaviour;

  //The messenger is used to send DomainGroupHandler messages
  DomainsFactory messenger;

  //Does the handling of dynamic groups
  private DynamicSubgroupsHandler dynamicSubgroupsHandler;

  private Integer coordination; //stores the coordinator or the member to be coordinated

  //The DomainMessagesHandler receives the messages to be processed (not the ST messages, those are handled here!)
  private DomainMessagesHandler messageHandler;

  //The DomainGroupHandler is required to send it on the acceptance event
  private GroupHandler groupHandler;
  private DomainGroupHandler domainGroupHandler;
  private DomainsUser domainsUser;

  //if not null, a thread is created doing the transder
  private StateSenderThread stateSenderThread;

  //Coordinations for the current transfer on the client side
  private CoordinationsMap coordinations;

  //set to to empty array if there is no propertyManager
  private MemberProperties [] nullMemberProperties;
  private Property[] nullProperties;

  //in case of inconsistencies on the group features, this field contains a valid reason for the expulsion
  private DomainExpulsionReason expulsionReason;

  //onBlockedView is true from the moment that a changeViewEvent is received until the new View Event is received
  private boolean onBlockedView;

  //keeps the state of this member. Only one of both can be non-null at one single moment in time
  private MemberInfo stateMemberInfo=null;
  private StatelessMemberInfo statelessMemberInfo=null;
}
