/**
  * File: DomainGroupHandlerImpl.java
  * Content: Implementation middleware-independent of the DomainGroupHandler interface.
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *  6th March 2002
  *    Note that there is no way to know if a member has been excluded or has not yet
  *    joined! The implementation should be done in such a way that after being
  *    expulsed, it is not possible to use the object
  *
  **/

package sensei.domains;
import sensei.middleware.domains.*;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.GroupMember;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;

import sensei.util.Error;

/**
  * Implementation middleware-independent of the DomainGroupHandler interface.
  *	It allows to define several subgroups to be handled separately by one groupHandler
  *	It includes the abstraction groupHandler/subgroups or domain/groups
  * It supports several special features:
  * -The sending of messages blocks the caller until the message is received
  * -Reception of cast messages sent by the own member are not propagated to the
  *		subgroups: they just unblock the message sending.
  * -PTP messages can be sent to all the subgroups in the groupHandler
  * -State transfer support
  * -Replica properties support
  *
  * To be done: cleaning after the member leaves the group. Some methods should additionally send
  *   InvalidGroupExceptions once that the member has been expulsed
  **/
public class DomainGroupHandlerImpl extends DomainGroupHandlerBaseImpl
{

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public DomainGroupHandlerImpl()
    throws Exception
  {
    Trace.code("DomainGroupHandlerImpl.java -> DomainGroupHandlerImpl ( )");
    this.user=new DomainsUser();
    this.behaviour=BehaviourOnViewChanges.StatelessMembersBelongToGroup;
    thisDomainGroupHandler = theDomainGroupHandler();
    groupHandler = null;
    behaviourImpl= null;
    subgroupsManager = new SubgroupsManager();
    eventsQueue = null;
    propertyManager = null;
    messageHandler = null;
    memberId = sensei.gms.Consts.generateValidMemberId();

    DomainsRegister.register(); //it can not be delayed until the behaviour construction
  }

  /**
   * After calling freeResources, the object becomes useless
   */
  public synchronized void freeResources()
  {
    Trace.code("DomainGroupHandlerImpl.java -> synchronized void freeResources ( )");
    if (user!=null)
    {
      clear();
      try
      {
        deactivate();
      }
      catch(Exception ex)
      {
        Trace.handledException(ex);
      }
      thisDomainGroupHandler=null;
      behaviour=null;
      subgroupsManager.freeResources();
      subgroupsManager=null;
      user=null;
    }
  }

  /**
   * Clear is called automatically at exclusion time, lets the object in the initial state
   */
  void clear()
  {
    Trace.code("DomainGroupHandlerImpl.java -> void clear ( )");
    user.clear();
    behaviour=BehaviourOnViewChanges.StatelessMembersBelongToGroup;
    groupHandler = null;
    if (behaviourImpl!=null)
    {
      behaviourImpl.freeResources();
      behaviourImpl= null;
    }
    subgroupsManager.clear();
    eventsQueue = null;
    if (propertyManager!=null)
    {
      propertyManager.freeResources();
      propertyManager = null;
    }
    if (messageHandler!=null)
    {
      messageHandler.freeResources();
      messageHandler = null;
    }
    dynamicSubgroupsHandler=null;
    memberId = sensei.gms.Consts.generateValidMemberId();

  }

  public void setBehaviourMode(BehaviourOnViewChanges mode) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setBehaviourMode ( BehaviourOnViewChanges mode )");
    checkMemberNotJoined();
    this.behaviour=mode;
  }

  public void setCoordinatorElector(CoordinatorElector elector) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setCoordinatorElector ( CoordinatorElector elector )");
    checkMemberNotJoined();
    user.setCoordinatorElector(elector);
  }

  public void setDomainGroupUser(DomainGroupUser dgUser) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setDomainGroupUser ( DomainGroupUser dgUser )");
    checkMemberNotJoined();
    user.setDomainGroupUser(dgUser);
  }

  public void setDynamicSubgroupsUser(DynamicSubgroupsUser subgroupsUser) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setDynamicSubgroupsUser ( DynamicSubgroupsUser subgroupsUser )");
    checkMemberNotJoined();
    user.setDynamicSubgroupsUser(subgroupsUser);
  }

  public void setPropertiesListener(PropertiesListener listener) throws PropertiesDisabledException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setPropertiesListener ( PropertiesListener listener )");
    checkPropertiesSupport();
    user.setPropertiesListener(listener);
  }

  public void enableProperties() throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void enableProperties ( )");
    checkMemberNotJoined();
    if (propertyManager==null) //no need to synchronize this, any problem would be harmless
      propertyManager = new PropertyManager();
  }

  public boolean arePropertiesEnabled()
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean arePropertiesEnabled ( )");
    return propertyManager!=null;
  }

  //*************************************************************************************//
  //**************************** GROUP MEMBER INTERFACE *********************************//
  //*************************************************************************************//

  /**
   * The member is accepted in the group, and receives its identity and the first view
   * It receives as well the group handler to use for communications
   **/
  public void memberAccepted(int id, GroupHandler handler, View view)
  {
    Trace.code("DomainGroupHandlerImpl.java -> void memberAccepted ( int id , GroupHandler handler , View view )");

    memberId = id;

    //When the member is accepted, the DomainGroupHandler starts really working
    if (propertyManager!=null)
      propertyManager.setMemberId(id);

    messageHandler = new DomainMessagesHandler (id, handler, subgroupsManager);
    //it is now created the BehaviourOnTransferImpl object, depending on the behaviour given
    switch(behaviour.value())
    {
      case BehaviourOnViewChanges._MembersOnTransferExcludedFromGroup:
        behaviourImpl = new MembersOnTransferExcludedFromGroupImpl(handler, thisDomainGroupHandler, user,
          subgroupsManager, propertyManager, messageHandler, id, view);
        break;
      case BehaviourOnViewChanges._StatelessMembersBelongToGroup:
        behaviourImpl = new StateMembersNotExcludedFromGroupImpl(handler, thisDomainGroupHandler, user,
          subgroupsManager, propertyManager, messageHandler, id, view, false);
          break;
      case BehaviourOnViewChanges._StatelessMembersDoNotBelongToGroup:
        behaviourImpl = new StateMembersNotExcludedFromGroupImpl(handler, thisDomainGroupHandler, user,
          subgroupsManager, propertyManager, messageHandler, id, view, true);
          break;
      default:
        Error.error(Consts.AREA,"DomainGroupHandlerImpl::memberAccepted");
    }

    eventsQueue = behaviourImpl.getEventsQueue();
    dynamicSubgroupsHandler = behaviourImpl.getDynamicSubgroupsHandler();

    synchronized(this)
    {
      groupHandler=handler; //accepted!!
    }

    behaviourImpl.memberAccepted(view);

    Trace.code("DomainGroupHandlerImpl.java -> void memberAccepted ENDED ( int id , GroupHandler handler , View view )");
  }

  /**
   * The member receives a point to point message to be processed
   * @param message The message to process
   **/
  public void processPTPMessage(int sender, Message msg)
  {
    Trace.code("DomainGroupHandlerImpl.java -> void processPTPMessage ( int sender , Message msg )");
    eventsQueue.newPTPMessage(sender, msg);
  }

  /**
   * The member receives a group message to be processed
   * @param message The message to process
   **/
  public void processCastMessage(int sender, Message msg)
  {
    Trace.code("DomainGroupHandlerImpl.java -> void processCastMessage ( int sender , Message msg )");
    eventsQueue.newCastMessage(sender, msg);
  }

  /**
   * The member receives a changing view event
   **/
  public void changingView()
  {
    Trace.code("DomainGroupHandlerImpl.java -> void changingView ( )");
    eventsQueue.changingView();
  }

  /**
   * The member receives a new view, including the list of members
   **/
  public void installView(View theView)
  {
    Trace.code("DomainGroupHandlerImpl.java -> void installView ( View theView )");
    eventsQueue.newView(theView);
  }

  /**
   * The member is excluded from the group, because it has requested it or because
   * it's considered to be faulty.
   **/
  public void excludedFromGroup()
  {
    Trace.code("DomainGroupHandlerImpl.java -> void excludedFromGroup ( )");
    eventsQueue.expulsed();
    //let the system like at starting point
    clear();
  }


  //*************************************************************************************//
  //**************************** GROUP HANDLER INTERFACE ********************************//
  //*************************************************************************************//
  /**
   * Returns the GroupMemberId handled by this GroupHandler
   * If the member is not active, the returned value is random - negative.
   **/
  public int getGroupMemberId()
  {
    Trace.code("DomainGroupHandlerImpl.java -> int getGroupMemberId ( )");
    return memberId;
  }

  /**
   * Returns true if the member belong to a group
   **/
  public boolean isValidGroup()
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean isValidGroup ( )");
    try
    {
      return isInGroup() && groupHandler.isValidGroup();
    }
    catch (Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
    return false;
  }

  /**
   * Leaves the group
   * @returns true if the member was belonging to a group
   **/
  public boolean leaveGroup()
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean leaveGroup ( )");
    boolean ret=behaviourImpl!=null;
    if (ret)
      try
      {
        ret=groupHandler.leaveGroup();
      }
      catch (Exception ex)
      {
        ret=false;
        Error.unhandledException(Consts.AREA, ex);
      }
    return ret;
  }

  /**
   * Sends a message to every member in the view, returning
   * true if the message is sent on this view
   * @param message The message to send
   * @returns true if the message is sent on the current view
   **/
  public boolean castMessage(Message msg)
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean castMessage ( Message msg )");
    return isInGroup()&&messageHandler.castMessage(msg);
  }

  /**
   * Sends a message to the specified member in the view, returning
   * true if the message is sent on this view. This call allows
   * to keep the order between cast and point-to-point messages
   * @param target The member receiving the message
   * @param message The message to send
   * @returns true if the message is sent on the current view
   **/
  public boolean sendMessage(int target, Message msg)
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean sendMessage ( int target , Message msg )");
    return isInGroup()&&messageHandler.sendMessage(target, msg);
  }

  //*************************************************************************************//
  //**************************** PROPERTY HANDLER INTERFACE *****************************//
  //*************************************************************************************//

  public MemberProperties[] getAllProperties() throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> MemberProperties [ ] getAllProperties ( )");
    checkPropertiesSupport();
    return propertyManager.getGroupProperties();
  }

  public MemberProperties[] getPropertyForAllMembers(String nam)
    throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> MemberProperties [ ] getPropertyForAllMembers ( String nam )");
    checkPropertiesSupport();
    return propertyManager.getPropertyForAllMembers(nam);
  }

  public Property[] getProperties() throws PropertiesDisabledException
  {
    Trace.code("DomainGroupHandlerImpl.java -> Property [ ] getProperties ( )");
    checkPropertiesSupport();
    return propertyManager.getProperties();
  }

  public String getProperty(String n) throws PropertiesDisabledException
  {
    Trace.code("DomainGroupHandlerImpl.java -> String getProperty ( String n )");
    checkPropertiesSupport();
    return propertyManager.getProperty(n);
  }

  public Property[] getMemberProperties(int loc) throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> Property [ ] getMemberProperties ( int loc )");
    checkPropertiesSupport();
    return propertyManager.getMemberProperties(loc);
  }

  public String getMemberProperty(int loc, String n) throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> String getMemberProperty ( int loc , String n )");
    checkPropertiesSupport();
    return propertyManager.getMemberProperty(loc, n);
  }

  public void setProperties(Property[] props) throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void setProperties ( Property [ ] props )");
    checkPropertiesSupport();
    checkMemberNotExpulsed();
    propertyManager.setProperties(props);
  }

  public void addProperties(Property[] props) throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void addProperties ( Property [ ] props )");
    checkPropertiesSupport();
    checkMemberNotExpulsed();
    propertyManager.addProperties(props);
  }

  public void removeProperties(Property[] props) throws PropertiesDisabledException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void removeProperties ( Property [ ] props )");
    checkPropertiesSupport();
    checkMemberNotExpulsed();
    propertyManager.removeProperties(props);
  }

  //*************************************************************************************//
  //**************************** TRANSACTIONS HANDLER INTERFACE *************************//
  //*************************************************************************************//

  public void startTransaction(Monitor monitor) throws MonitorException, TransactionException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void startTransaction ( Monitor monitor )");
    checkInGroup(true);
    behaviourImpl.startTransaction(monitor);
  }

  public void endTransaction() throws MonitorException, TransactionException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void endTransaction ( )");
    checkInGroup(true);
    behaviourImpl.endTransaction();
  }

  //*************************************************************************************//
  //**************************** DOMAIN GROUP HANDLER INTERFACE *************************//
  //*************************************************************************************//

  public int[] getStatefulMembers() throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> int [ ] getStatefulMembers ( )");
    if (behaviourImpl==null)
      throw new MemberStateException(MemberStateExceptionReason.MemberNotJoined);
    return behaviourImpl.getStatefulMembers();
  }

  public void registerSubgroup(int uniqueSubgroupId, GroupMember subgroup)
    throws MemberStateException, SubgroupsHandlerException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void registerSubgroup ( int uniqueSubgroupId , GroupMember subgroup )");
    checkInGroup(false); //cannot be still accepted
    if ((uniqueSubgroupId > MAX_STATIC_SUBGROUP_ID) || (uniqueSubgroupId == Consts.EVERY_SUBGROUP))
      throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.InvalidStaticSubgroupId);
    try
    {
      subgroupsManager.registerSubgroup(uniqueSubgroupId,STGroupMember.createSTGroupMember(subgroup),staticSubgroupInfo);
    }
    catch(SubgroupException ex)
    {
      //not possible on static groups!
      Error.unhandledException(Consts.AREA, ex);
    }
  }

  public void castSubgroupCreation(DynamicSubgroupInfo info)
    throws SubgroupsHandlerException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void castSubgroupCreation ( DynamicSubgroupInfo info )");
    checkInGroup(true); //must be accepted
    dynamicSubgroupsHandler.castSubgroupCreation(info);
  }

  public int createSubgroup(DynamicSubgroupInfo info, GroupMember subgroup)
    throws SubgroupsHandlerException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> int createSubgroup ( DynamicSubgroupInfo info , GroupMember subgroup )");
    checkInGroup(true); //must be accepted
    return dynamicSubgroupsHandler.createSubgroup(info, subgroup);
  }

  public boolean removeSubgroup(int id, DynamicSubgroupInfo info)
    throws SubgroupsHandlerException, MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean removeSubgroup ( int id , DynamicSubgroupInfo info )");
    checkInGroup(true); //must be accepted
    return dynamicSubgroupsHandler.removeSubgroup(id, info);
  }

  public int[] getSubgroups()
  {
    Trace.code("DomainGroupHandlerImpl.java -> int [ ] getSubgroups ( )");
    return subgroupsManager.getSubgroups();
  }

   /**
    * Returns the subgroup associated to a given subgroupId, null if not existing
    */
  public GroupMember getSubgroup(int id)
  {
    Trace.code("DomainGroupHandlerImpl.java -> GroupMember getSubgroup ( int id )");
    return subgroupsManager.getSubgroup(id);
  }

   /**
    * Returns the subgroupId for a group. If the group is not valid, EVERY_SUBGROUP is returned
    */
  public int getSubgroupId(GroupMember subgroup)
  {
    Trace.code("DomainGroupHandlerImpl.java -> int getSubgroupId ( GroupMember subgroup )");
    return subgroupsManager.getSubgroupId(subgroup);
  }

  /**
    * Cast a message to the subgroup, and blocks until it is processed.
    **/
  public boolean syncCastDomainMessage(DomainMessage message, boolean normalProcessing) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> boolean syncCastDomainMessage ( DomainMessage message , boolean normalProcessing )");
    checkInGroup(true);
    return messageHandler.syncCastDomainMessage(message, normalProcessing);
  }

  public void syncCastDomainMessageProcessed() throws MemberStateException, SyncCastDomainMessageProcessedException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void syncCastDomainMessageProcessed ( )");
    checkInGroup(true);
    messageHandler.syncCastDomainMessageProcessed();
  }

  //*************************************************************************************//
  //**************************** PRIVATE METHODS ****************************************//
  //*************************************************************************************//

  boolean isInGroup(){return groupHandler!=null;}
  void checkInGroup(boolean is) throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void checkInGroup ( boolean is )");
    if (isInGroup()!=is)
    {
      MemberStateExceptionReason reason;
      if (is) //can be not joined or expulsed
        if (eventsQueue==null) //expulsed
          reason = MemberStateExceptionReason.MemberExcluded;
        else
          reason = MemberStateExceptionReason.MemberNotJoined;
      else
        reason = MemberStateExceptionReason.MemberJoined;
      throw new MemberStateException(reason);
    }
  }
  public void checkMemberNotJoined() throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void checkMemberNotJoined ( )");
    if (behaviourImpl!=null)
      throw new MemberStateException(MemberStateExceptionReason.MemberJoined);
  }
  void checkMemberNotExpulsed() throws MemberStateException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void checkMemberNotExpulsed ( )");
    //Note that there is no way to know if a member has been excluded or has not yet
    // joined! The implementation should be done in such a way that after being
    // expulsed, it is not possible to use the object
    //    if (eventsQueue==null)
    //      throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);

  }

  /**
   * Throws an exception if properties are not supported
   */
  public void checkPropertiesSupport() throws PropertiesDisabledException
  {
    Trace.code("DomainGroupHandlerImpl.java -> void checkPropertiesSupport ( )");
    if (propertyManager==null)
      throw new PropertiesDisabledException();
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  int memberId;

  //thisDomainGroupHandler is the published instance of this object
  DomainGroupHandler thisDomainGroupHandler;

  //there is a GroupHandler behalf the DomainGroupHandler, null when it is not yet activated
  GroupHandler groupHandler;

  //The behaviour on StateTransfer is delegated into BehaviourOnTransferImpl instances
  //The exact BehaviourOnTransferImpl to use depends on the behaviour, set on setFeatures
  BehaviourOnTransferImpl behaviourImpl;
  BehaviourOnViewChanges behaviour;

  //The subgroups are handled through a SubgroupsManager class
  SubgroupsManager subgroupsManager;

  //Properties handler
  PropertyManager propertyManager;

  //Any event is automatically dispatched to the EventsQueue object, which controls the
  //flow asynchronously to the EventsQueueUser (a BehaviourTransferImpl object)
  //It is only created once that there are events to handle (after memberIsAccepted), and it
  //is created on the BehaviourOnTransferImpl object
  EventsQueue eventsQueue;

  //Does the handling of dynamic groups
  DynamicSubgroupsHandler dynamicSubgroupsHandler;

  //The user of the application
  DomainsUser user;

  //The DomainMessagesHandler manages the cast message operations, and it is used directly by
  //the BehaviourImplementation to deliver the unqueued messages
  DomainMessagesHandler messageHandler;

  final static int MAX_STATIC_SUBGROUP_ID = sensei.middleware.domains.MAX_STATIC_SUBGROUP_ID.value;
  final static DynamicSubgroupInfo staticSubgroupInfo = new DynamicSubgroupInfoImpl();
}
