/**
  * File: DomainsUser.java
  * Content: Instance centralizing all the different user types in the domains module
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupMember;
import sensei.middleware.domains.CoordinatorElector;
import sensei.middleware.domains.CoordinatorInformation;
import sensei.middleware.domains.DomainExpulsionReason;
import sensei.middleware.domains.DomainGroupUser;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.DynamicSubgroupsUser;
import sensei.middleware.domains.PropertiesDisabledException;
import sensei.middleware.domains.PropertiesListener;
import sensei.middleware.domains.State;
import sensei.middleware.domains.StateTransferType;

import sensei.util.Error;

/**
 * Instance centralizing all the different user types in the domains module
  **/
class DomainsUser
{
  //*************************************************************************************
  //**************************** CONSTRUCTOR ********************************************
  //*************************************************************************************

  public DomainsUser()
  {
    Trace.code("DomainsUser.java -> DomainsUser ( )");
    clear();
  }

  public void clear()
  {
    Trace.code("DomainsUser.java -> void clear ( )");
    dgUser=null;
    subgroupsUser=null;
    coordinatorElector=null;
    propertiesListener=null;
  }

  public void setDomainGroupUser(DomainGroupUser user)
  {
    Trace.code("DomainsUser.java -> void setDomainGroupUser ( DomainGroupUser user )");
    dgUser = user;
  }

  public void setDynamicSubgroupsUser(DynamicSubgroupsUser user)
  {
    Trace.code("DomainsUser.java -> void setDynamicSubgroupsUser ( DynamicSubgroupsUser user )");
    subgroupsUser=user;
  }

  public void setCoordinatorElector(CoordinatorElector elector)
  {
    Trace.code("DomainsUser.java -> void setCoordinatorElector ( CoordinatorElector elector )");
    coordinatorElector=elector;
  }

  public void setPropertiesListener(PropertiesListener listener)
  {
    Trace.code("DomainsUser.java -> void setPropertiesListener ( PropertiesListener listener )");
    propertiesListener=listener;
  }

  //*************************************************************************************
  //**************************** DOMAIN GROUP USER **************************************
  //*************************************************************************************

  public void domainAccepted(int memberId)
  {
    Trace.code("DomainsUser.java -> void domainAccepted ( int memberId )");
    DomainGroupUser user = dgUser;
    if (user!=null)
      try
      {
        user.domainAccepted(memberId);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }
  }

  public void domainExpulsed(DomainExpulsionReason reason)
  {
    Trace.code("DomainsUser.java -> void domainExpulsed ( DomainExpulsionReason reason )");
    DomainGroupUser user = dgUser;
    if (user!=null)
      try
      {
        user.domainExpulsed(reason);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }
  }

  public void stateObtained(boolean assumed)
  {
    Trace.code("DomainsUser.java -> void stateObtained ( boolean assumed )");
    DomainGroupUser user = dgUser;
    if (user!=null)
      try
      {
        user.stateObtained(assumed);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }
  }

  public void offendingSubgroup(int subGroupId, String reason)
  {
    Trace.code("DomainsUser.java -> void offendingSubgroup ( int subGroupId , String reason )");
    DomainGroupUser user = dgUser;
    if (user!=null)
      try
      {
        user.offendingSubgroup(subGroupId, reason);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }
  }

  //*************************************************************************************
  //**************************** COORDINATOR ELECTOR ************************************
  //*************************************************************************************

  public boolean supportsCoordinatorElection()
  {
    Trace.code("DomainsUser.java -> boolean supportsCoordinatorElection ( )");
    return coordinatorElector!=null;
  }

  /**
   * Only valid if there is coordinator election support
   */
  public int getCoordinator (int loc, CoordinatorInformation[] info)
  {
    Trace.code("DomainsUser.java -> int getCoordinator ( int loc , CoordinatorInformation [ ] info )");
    CoordinatorElector elector = coordinatorElector;

    assert elector!=null;

    int ret = Consts.INVALID_MEMBER_ID;

    try
    {
      ret = elector.getCoordinator(loc, info);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }

    return ret;
  }

  //*************************************************************************************
  //**************************** PROPERTIES LISTENER ************************************
  //*************************************************************************************

  public void propertiesUpdated (int member)
  {
    Trace.code("DomainsUser.java -> void propertiesUpdated ( int member )");
    PropertiesListener listener = propertiesListener;

    if (listener!=null)
      try
      {
        listener.propertiesUpdated(member);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex);
      }

  }

  //*************************************************************************************
  //**************************** DYNAMIC SUBGROUPS USER *********************************
  //*************************************************************************************

  public boolean supportsDynamicBehaviour(){return subgroupsUser!=null;}

  public GroupMember acceptSubgroup(int subgroupId,  DynamicSubgroupInfo info)
  {
    Trace.code("DomainsUser.java -> GroupMember acceptSubgroup ( int subgroupId , DynamicSubgroupInfo info )");
    DynamicSubgroupsUser user = subgroupsUser;

    assert user!=null;

    GroupMember ret = null;

    try
    {
      ret = user.acceptSubgroup(subgroupId, info);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }

    return ret;
  }

  public GroupMember subgroupCreated(int creator, int subgroupId,  DynamicSubgroupInfo info)
  {
    Trace.code("DomainsUser.java -> GroupMember subgroupCreated ( int creator , int subgroupId , DynamicSubgroupInfo info )");
    DynamicSubgroupsUser user = subgroupsUser;

    assert user!=null;

    GroupMember ret = null;

    try
    {
      ret = user.subgroupCreated(creator, subgroupId, info);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }

    return ret;
  }

  public void subgroupRemoved(int remover, int subgroupId, DynamicSubgroupInfo info)
  {
    Trace.code("DomainsUser.java -> void subgroupRemoved ( int remover , int subgroupId , DynamicSubgroupInfo info )");
    DynamicSubgroupsUser user = subgroupsUser;

    assert user!=null;

    try
    {
      user.subgroupRemoved(remover, subgroupId, info);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex);
    }
  }


  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  DomainGroupUser dgUser;
  DynamicSubgroupsUser subgroupsUser;
  CoordinatorElector coordinatorElector;
  PropertiesListener propertiesListener;
}
