/**
  * File: DynamicSubgroupsHandler.java
  * Content: handles the creation of dynamic subgroups
  * Author: LuisM Pena
  * Date: 24th October 2001
  * Version: 0.51.00
  * Last change:
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupMember;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MemberStateExceptionReason;
import sensei.middleware.domains.MessageCreateSubgroup;
import sensei.middleware.domains.MessageRemoveSubgroup;
import sensei.middleware.domains.State;
import sensei.middleware.domains.StateTransferType;
import sensei.middleware.domains.SubgroupsHandlerException;
import sensei.middleware.domains.SubgroupsHandlerExceptionReason;
import sensei.middleware.domains.SubgroupProperties;

import sensei.util.Error;
import sensei.util.IntWrapper;

import java.util.HashMap;
import java.util.Map;

/**
 * handles the creation of dynamic subgroups
 **/
class DynamicSubgroupsHandler
{
  //*************************************************************************************
  //**************************** CONSTRUCTOR ********************************************
  //*************************************************************************************

  public DynamicSubgroupsHandler(int thisMemberId, DomainsUser user, SubgroupsManager subgroupsManager,
    DomainsFactory messenger)
  {
    Trace.code("DynamicSubgroupsHandler.java -> DynamicSubgroupsHandler ( int thisMemberId , DomainsUser user , SubgroupsManager subgroupsManager , DomainsFactory messenger )");
    this.messenger=messenger;
    this.thisMemberId = thisMemberId;
    this.user=user;
    this.subgroupsManager=subgroupsManager;
    subgroupsBeingCreated = new HashMap();
    wrapper = new IntWrapper();
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public void freeResources()
  {
    Trace.code("DynamicSubgroupsHandler.java -> void freeResources ( )");
    if (messenger!=null)
    {
      messenger=null;
      wrapper=null;
      user=null;
      synchronized(this)
      {
        subgroupsBeingCreated.clear();
        notifyAll();
      }
      subgroupsManager=null;
    }
  }

  //*************************************************************************************//
  //**************************** GET NEXT DYNAMIC SUBGROUP ID ***************************//
  //*************************************************************************************//

  /**
   * Returns the next dynamic subgroup id to be used
   */
  public int getNextDynamicSubgroupId()
  {
    Trace.code("DynamicSubgroupsHandler.java -> int getNextDynamicSubgroupId ( )");
    assert user!=null;
    return user.supportsDynamicBehaviour()? nextSubgroupId : Consts.EVERY_SUBGROUP;
  }

  //*************************************************************************************//
  //**************************** CREATE  DYNAMIC SUBGROUPS ******************************//
  //*************************************************************************************//

  /**
   * Creates the given dynamic subgroups. Returns false if any of the created groups does not respect the
   * expected type
   * @exception SubgroupException after an exception accessing any of the created subgroups
   */
  public boolean createDynamicSubgroups(int nextDynamicSubgroupId, SubgroupProperties subgroupProperties[])
    throws SubgroupException
  {
    Trace.code("DynamicSubgroupsHandler.java -> boolean createDynamicSubgroups ( int nextDynamicSubgroupId , SubgroupProperties subgroupProperties [ ] )");
    assert user!=null;
    assert user.supportsDynamicBehaviour();

    //perhaps another subgroup has been created on the meanwhile!
    nextSubgroupId = Math.max(nextSubgroupId, nextDynamicSubgroupId);

    boolean ok=true;
    int i = subgroupProperties.length;
    while(ok && (i-->0))
    {
      SubgroupProperties subgroup = subgroupProperties[i];
      if ((subgroup.id > Consts.MAX_STATIC_SUBGROUP_ID) && (!subgroupsManager.containsSubgroup(subgroup.id)))
        ok = createDynamicSubgroup(subgroup.id, subgroup.info, subgroup.type);
    }
    return ok;
  }


  /**
   * Returns false if the state transfer type is not valid. An exception can be raised when the subgroup
   * is accessed
   */
  boolean createDynamicSubgroup(int id, DynamicSubgroupInfo info, StateTransferType type) throws SubgroupException
  {
    Trace.code("DynamicSubgroupsHandler.java -> boolean createDynamicSubgroup ( int id , DynamicSubgroupInfo info , StateTransferType type )");
    STGroupMember subgroup = STGroupMember.createSTGroupMember(user.acceptSubgroup(id, info));
    boolean valid = subgroup.getStateTransferType()==type;
    if (valid)
      try
      {
        subgroupsManager.registerSubgroup(id, subgroup, info);
      }
      catch(SubgroupsHandlerException sex)
      {
        Error.unhandledException(Consts.AREA, sex);
      }
    return valid;
  }

  //*************************************************************************************
  //**************************** CREATE SUBGROUP ****************************************
  //*************************************************************************************

  public void castSubgroupCreation(DynamicSubgroupInfo info) throws SubgroupsHandlerException
  {
    Trace.code("DynamicSubgroupsHandler.java -> void castSubgroupCreation ( DynamicSubgroupInfo info )");
    assert user!=null;
    if (!user.supportsDynamicBehaviour())
      throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.DynamicBehaviourNotRegistered);

    try
    {
      messenger.castMessageCreateSubgroup(info, 0, false);
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
    }
  }

  public int createSubgroup(DynamicSubgroupInfo info, GroupMember subgroup)
    throws SubgroupsHandlerException, MemberStateException
  {
    Trace.code("DynamicSubgroupsHandler.java -> int createSubgroup ( DynamicSubgroupInfo info , GroupMember subgroup )");
    assert user!=null;
    if (!user.supportsDynamicBehaviour())
      throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.DynamicBehaviourNotRegistered);

    int ret = Consts.EVERY_SUBGROUP;
    try
    {
      IntWrapper subgroupId = new IntWrapper(ret);
      STGroupMember toCreate = STGroupMember.createSTGroupMember(subgroup);
      synchronized (this)
      {
        //send the message and wait for the answer
        subgroupId.i = --creatorId;
        subgroupsBeingCreated.put(subgroupId, toCreate);
      }
      messenger.castMessageCreateSubgroup(info, subgroupId.i, true);

      //in the map, is inserted the group to create; when is created, it is substitued by is id
      while(true)
      {
        synchronized(this)
        {
          if (subgroupsBeingCreated.isEmpty())
            throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);
          else
          {
            Object got = subgroupsBeingCreated.get(subgroupId);
            if (got==toCreate)
              try{wait();}catch(InterruptedException ex){}
            else
            {
              ret=((IntWrapper)got).i;
              subgroupsBeingCreated.remove(subgroupId);
              break;
            }
          }
        }
      }
    }
    catch(MemberStateException ex){throw ex;}
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE CREATE SUBGROUP ************************
  //*************************************************************************************

  /**
   * Returns false if there is a problem on state transfer types creating the subgroup
   * @exception SubgroupException if there are problems accesing the created subgroup
   */
  public boolean processMessageCreateSubgroup(int sender, MessageCreateSubgroup message) throws SubgroupException
  {
    Trace.code("DynamicSubgroupsHandler.java -> boolean processMessageCreateSubgroup ( int sender , MessageCreateSubgroup message )");
    assert user!=null;
    boolean ret=true;
    if (user.supportsDynamicBehaviour())
    {
      IntWrapper createdSubgroupId = new IntWrapper(nextSubgroupId++);
      STGroupMember subgroup = null;
      boolean notify=false;

      //notify to the user that it must be created, but only if the sender is not this member and it
      //has already supplied the subgroup
      if ((sender == thisMemberId) && message.waitReception)
      {
        wrapper.i = message.creatorId;
        subgroup = (STGroupMember) subgroupsBeingCreated.put(wrapper, createdSubgroupId);
        assert subgroup!=null;
        notify=true;
      }
      else
      {
        subgroup = STGroupMember.createSTGroupMember
            (user.subgroupCreated(sender, createdSubgroupId.i, message.info));
      }

      try
      {
        if (ret)
          subgroupsManager.registerSubgroup(createdSubgroupId, subgroup, message.info);
        if (notify)
          synchronized(this){notifyAll();}
      }
      catch (SubgroupException ex){throw ex;}
      catch(Exception ex){Error.unhandledException(Consts.AREA, ex);}
    }
    else
    {
      Trace.errorReceivingMessageDynamicBehaviour();
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** REMOVE SUBGROUP ****************************************
  //*************************************************************************************

  public boolean removeSubgroup(int subgroupId, DynamicSubgroupInfo info) throws SubgroupsHandlerException
  {
    Trace.code("DynamicSubgroupsHandler.java -> boolean removeSubgroup ( int subgroupId , DynamicSubgroupInfo info )");
    assert user!=null;
    if (!user.supportsDynamicBehaviour())
      throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.DynamicBehaviourNotRegistered);

    if (subgroupId<=Consts.MAX_STATIC_SUBGROUP_ID)
      throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.InvalidDynamicSubgroupId);

    boolean ret = subgroupsManager.containsSubgroup(subgroupId);
    if (ret)
    {
      try
      {
        messenger.castMessageRemoveSubgroup(subgroupId, info);
      }
      catch(Exception ex)
      {
        Error.unhandledException(Consts.AREA, ex); //Remote/CORBA exception, menaingless as the access is local!!!
      }
    }
    return ret;
  }

  //*************************************************************************************
  //**************************** PROCESS MESSAGE REMOVE SUBGROUP ************************
  //*************************************************************************************

  public void processMessageRemoveSubgroup(int sender, MessageRemoveSubgroup message) throws SubgroupException
  {
    Trace.code("DynamicSubgroupsHandler.java -> void processMessageRemoveSubgroup ( int sender , MessageRemoveSubgroup message )");
    assert user!=null;
    if (user.supportsDynamicBehaviour())
    {
      user.subgroupRemoved(sender, message.subgroupToRemove, message.info);

      subgroupsManager.unregisterSubgroup(message.subgroupToRemove);
    }
    else
    {
      Trace.errorReceivingMessageDynamicBehaviour();
    }
  }

  //*************************************************************************************
  //**************************** DATA MEMBERS *******************************************
  //*************************************************************************************

  DomainsFactory messenger;
  DomainsUser user;
  Map subgroupsBeingCreated; //IntWrapper, {STGroupMember other IntWrapper}
  SubgroupsManager subgroupsManager;
  IntWrapper wrapper;
  int thisMemberId;
  int nextSubgroupId = Consts.MAX_STATIC_SUBGROUP_ID + 1;
  int creatorId = 0;
}
