/**
  * File: MembersStateHandler.java
  * Content: Class controlling the state of the transfer for the members in the group
  * Author: LuisM Pena
  * Date: 4th May 2001
  * Version: 0.51.00
  * Last change:
  *
  **/

package sensei.domains;

import sensei.middleware.domains.CoordinatorInformation;
import sensei.middleware.domains.ReplicaInformation;

import sensei.util.Error;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
  * Class controlling the state of the transfer for the members in the group
  **/
class MembersStateHandler
{

  class OnQueue
  {
    public OnQueue(Integer coordinator, Integer newMember)
    {
      this.coordinator=coordinator;
      this.newMember=newMember;
    }
    Integer coordinator;
    Integer newMember;
  };
  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public MembersStateHandler(int viewId, int members[], int newMembers[], DomainsUser user)
  {
    Trace.code("MembersStateHandler.java -> MembersStateHandler ( int viewId , int members [ ] , int newMembers [ ] , DomainsUser user )");
    stateMembers = new TreeMap();
    statelessMembers = new HashMap();
    domainsUser = user;
    membersToCoordinate = new TreeSet();
    queuedStateMembers = new ArrayList();
    informationBuilt=false;

    //introduce the members information
    addMembers(members, 0); //for the existing members, it does not mind the view, as far as is low enough
    addMembers(newMembers, viewId);
  }

  //*************************************************************************************
  //**************************** FREE RESOURCES *****************************************
  //*************************************************************************************
  /**
   * Frees any resources, the object can not be used anymore
   */
  public synchronized void freeResources()
  {
    Trace.code("MembersStateHandler.java -> synchronized void freeResources ( )");
    if (stateMembers!=null)
    {
      stateMembers.clear();
      stateMembers=null;
      statelessMembers.clear();
      statelessMembers=null;
      queuedStateMembers.clear();
      queuedStateMembers=null;
      domainsUser=null;
      membersToCoordinate.clear();
      membersToCoordinate=null;
    }
  }

  //*************************************************************************************//
  //**************************** ASSUME STATE *******************************************//
  //*************************************************************************************//

  /**
   * The specified member is given state condition, moving it from the stateteless list
   * to the state one.
   * The rebuildInformation cannot have been called.
   */
  public synchronized void assumeState(Integer member)
  {
    Trace.code("MembersStateHandler.java -> synchronized void assumeState ( Integer member )");
    assert stateMembers!=null;
    assert !informationBuilt;

    Object dbgHelp = statelessMembers.remove(member);
    assert dbgHelp!=null;

    stateMembers.put(member, new MemberInfo());

    assert membersToCoordinate.isEmpty();
    membersToCoordinate.addAll(statelessMembers.keySet());
    assignCoordinator(membersToCoordinate);
    membersToCoordinate.clear();

    informationBuilt=true;
  }

  //*************************************************************************************//
  //**************************** REBUILD INFORMATION ************************************//
  //*************************************************************************************//

  /**
   * The MembersStateHandler builds its information from the transfer from another member
   * (on the message p)
   */
  public synchronized void rebuildInformation(ReplicaInformation[] replicasInfo)
  {
    Trace.code("MembersStateHandler.java -> synchronized void rebuildInformation ( ReplicaInformation [ ] replicasInfo )");
    assert stateMembers!=null;
    assert !informationBuilt;
    assert stateMembers.isEmpty();
    int len = replicasInfo.length;
    for (int i=0;i <len; i++)
    {
      ReplicaInformation replicaInfo = replicasInfo[i];
      Integer member = new Integer(replicaInfo.replicaId);

      //if the member is not in the statelessMembers, cannot be passed to the state members!
      Object dbgHelp = statelessMembers.remove(member);
      assert dbgHelp!=null;

      //put it as stateMember
      MemberInfo info = new MemberInfo();
      dbgHelp = stateMembers.put(member, info);
      assert dbgHelp==null;

      //The coordinating members information must be transfered to both, the StateMember doing the coordination
      //and the stateless one that receives it
      int coord[] = replicaInfo.replicasToCoordinate;
      int len2=coord.length;
      for(int j=0;j<len2;j++)
      {
        Integer wrapper = new Integer(coord[j]);
        StatelessMemberInfo statelessMemberInfo = (StatelessMemberInfo) (statelessMembers.get(wrapper));
        assert statelessMemberInfo!=null;
        assert statelessMemberInfo.getCoordinator()==null;

        info.addMemberToCoordinate(wrapper);
        statelessMemberInfo.setCoordinator(member);
      }
    }

    informationBuilt=true;

    //There could have been some members which have already sent a message E saying that they have
    //achieved state. Those messages should be queued and processed now
    if (!queuedStateMembers.isEmpty())
    {
      Iterator iterator=queuedStateMembers.iterator();
      while (iterator.hasNext())
      {
        OnQueue queued = (OnQueue) (iterator.next());
        memberGotStateFromCoordinator(queued.coordinator, queued.newMember);
      }
      queuedStateMembers.clear();
    }
  }

  //*************************************************************************************//
  //**************************** CLEAR INFORMATION **************************************//
  //*************************************************************************************//

  /**
   * Removes the stateMembers structure, passing those members to the stateless structure.
   * From this last structure, the coordinators are also removed
   */
  public synchronized void clearInformation()
  {
    Trace.code("MembersStateHandler.java -> synchronized void clearInformation ( )");
    assert stateMembers!=null;
    Iterator iterator = stateMembers.keySet().iterator();
    while (iterator.hasNext())
    {
      Integer member = (Integer)(iterator.next());
      Object previous = statelessMembers.put(member, new StatelessMemberInfo(0));
      assert previous==null;
    }
    stateMembers.clear();
    iterator=statelessMembers.values().iterator();
    while (iterator.hasNext())
    {
      StatelessMemberInfo info = (StatelessMemberInfo) (iterator.next());
      info.setCoordinator(null);
    }
    queuedStateMembers.clear();
    informationBuilt=false;
  }

  //*************************************************************************************//
  //**************************** MEMBER GOT STATE FROM COORDINATOR **********************//
  //*************************************************************************************//

  /**
   * Moves a member from the statelessMember group to the statefull one, returning false
   * if the operation is not possible for any reason.
   * This operation is mean to be invoked after a valid messageE
   */
  public synchronized boolean memberGotStateFromCoordinator(Integer coordinator, Integer newMember)
  {
    Trace.code("MembersStateHandler.java -> synchronized boolean memberGotStateFromCoordinator ( Integer coordinator , Integer newMember )");
    assert stateMembers!=null;
    boolean ret=false;
    if (informationBuilt)
    {
      MemberInfo info = (MemberInfo) (stateMembers.get(coordinator));
      if (info!=null) //the coordinator must be a statefull member.
      {
        Integer coordinated = info.getMemberToCoordinate();
        if ((coordinated!=null) && coordinated.equals(newMember)) //it is checked that the member is indeed being coordinated
        {
          Object removed=statelessMembers.remove(newMember); //remove the object as stateless member
          assert removed!=null;

          info.removeMemberToCoordinate(coordinated); //remove the member from the list in the coordinator

          MemberInfo newInfo = new MemberInfo(); //create the information for the new member

          Object reinserted=stateMembers.put(coordinated, newInfo);
          assert reinserted==null;

          ret=true;
        }
      }
    }
    else
    {
      //the message P has not been received, this request must be queued and be processed later
      queuedStateMembers.add(new OnQueue(coordinator, newMember));
    }
    return ret;
  }

  //*************************************************************************************//
  //**************************** ADD MEMBERS ********************************************//
  //*************************************************************************************//

  public synchronized void addMembers(int members[], int viewId)
  {
    Trace.code("MembersStateHandler.java -> synchronized void addMembers ( int members [ ] , int viewId )");
    assert stateMembers!=null;
    assert membersToCoordinate.size()==0;

    int len = members.length;
    while(len-->0)
    {
      Integer wrapper = new Integer(members[len]);
      membersToCoordinate.add(wrapper); //if it was already present, it's not re-added!
      statelessMembers.put(wrapper, new StatelessMemberInfo(viewId)); //if was already present, is just changed
    }

    assignCoordinator(membersToCoordinate);
    membersToCoordinate.clear();
  }

  //*************************************************************************************//
  //**************************** REMOVE MEMBERS *****************************************//
  //*************************************************************************************//

  /**
   * The members specified are removed.
   * Removing a stateless member means to look for its coordinator and remove as well that dependency
   * Removing a state member means to reassign any member being coordinated
   */
  public synchronized void removeMembers(int members[])
  {
    Trace.code("MembersStateHandler.java -> synchronized void removeMembers ( int members [ ] )");
    assert stateMembers!=null;
    assert membersToCoordinate.size()==0;

    int len = members.length;
    while(len-->0)
    {
      Integer wrapper = new Integer(members[len]);

      //look for the member as a statefull one
      MemberInfo info = (MemberInfo) (stateMembers.remove(wrapper));
      if (info==null)
      {
        //not a statefull, it must be then a member without state
        StatelessMemberInfo saved = (StatelessMemberInfo) (statelessMembers.remove(wrapper));
        assert saved!=null;

        //inform to its coordinator that it doesn't have to do this coordination
        Integer coordinator = saved.getCoordinator();
        if (coordinator!=null)
        {
          //perhaps the coordinator is also down!
          info = (MemberInfo) (stateMembers.get(coordinator));
          if (info == null)
          {
            //the coordinator has been removed, remove it now from the list of membersToCoordinate
            boolean contained = membersToCoordinate.remove(wrapper);
            assert contained;
          }
          else
          {
            boolean contained = info.removeMemberToCoordinate(wrapper);
            assert contained;
          }
        }
      }
      else
      {
        //if the member is a statefull, its members to coordinate must be redistributed
        info.getCoordinatedMembers(membersToCoordinate);
      }
    }
    assignCoordinator(membersToCoordinate);
    membersToCoordinate.clear();
  }

  //*************************************************************************************//
  //**************************** HAS STATE THE MEMBER ***********************************//
  //*************************************************************************************//

  public synchronized boolean hasStateTheMember(Integer member)
  {
    Trace.code("MembersStateHandler.java -> synchronized boolean hasStateTheMember ( Integer member )");
    assert stateMembers!=null;
    return stateMembers.containsKey(member);
  }

  //*************************************************************************************//
  //**************************** GET STATE INFO *****************************************//
  //*************************************************************************************//

  /**
   * Returns the MemberInfo associated to a state member, or null (and asserted) if it is not
   * a state member
   */
  public synchronized MemberInfo getStateInfo(Integer member)
  {
    Trace.code("MembersStateHandler.java -> synchronized MemberInfo getStateInfo ( Integer member )");
    assert stateMembers!=null;
    MemberInfo ret = (MemberInfo) (stateMembers.get(member));
    assert ret!=null;
    return ret;
  }

  //*************************************************************************************//
  //**************************** GET STATELESS INFO *************************************//
  //*************************************************************************************//

  /**
   * Returns the StatelessMemberInfo associated to a stateless member, or null (and asserted) if it is not
   * a state member
   */
  public synchronized StatelessMemberInfo getStatelessInfo(Integer member)
  {
    Trace.code("MembersStateHandler.java -> synchronized StatelessMemberInfo getStatelessInfo ( Integer member )");
    assert stateMembers!=null;
    StatelessMemberInfo ret = (StatelessMemberInfo)(statelessMembers.get(member));
    assert ret!=null;
    return ret;
  }

  //*************************************************************************************//
  //**************************** GET OLDEST STATELESS MEMBER ****************************//
  //*************************************************************************************//

  /**
    * Return the oldest stateless member, or null if there are no stateless members or there are state members
    * If two members share the same initialViewId, the lower groupMemberId determines
    * the younger member
    **/
  public synchronized Integer getOldestStatelessMember()
  {
    Trace.code("MembersStateHandler.java -> synchronized Integer getOldestStatelessMember ( )");
    assert stateMembers!=null;
    Integer ret=null;
    if (stateMembers.isEmpty())
    {
      int itsView=0;
      Iterator iterator = statelessMembers.entrySet().iterator();
      while(iterator.hasNext())
      {
        Map.Entry entry = (Map.Entry) (iterator.next());
        Integer member = (Integer)(entry.getKey());
        int view = ((StatelessMemberInfo) (entry.getValue())).getInitialView();
        if ((ret==null) || (itsView>view) || ((itsView==view) && (member.compareTo(ret)<0)))
        {
          ret=member;
          itsView=view;
        }
      }
    }
    return ret;
  }

  //*************************************************************************************//
  //**************************** GET MEMBERS ********************************************//
  //*************************************************************************************//

  /**
   * Returns the members (the stateful members, therefore)
   */
  public synchronized int[] getMembers()
  {
    Trace.code("MembersStateHandler.java -> synchronized int [ ] getMembers ( )");
    assert stateMembers!=null;
    int size = stateMembers.size();
    int[] ret=new int[size];
    int index=0;
    Iterator iterator = stateMembers.entrySet().iterator();
    while (iterator.hasNext())
    {
      assert index<size;
      Map.Entry entry = (Map.Entry)(iterator.next());
      Integer replica = (Integer)(entry.getKey());
      ret[index++]=replica.intValue();
    }
    assert index==size;
    return ret;
  }

  //*************************************************************************************//
  //**************************** GET REPLICAS INFORMATION *******************************//
  //*************************************************************************************//

  /**
   * Retrieves the replicas information, as it defined on the message P
   */
  public synchronized ReplicaInformation[] getReplicasInformation()
  {
    Trace.code("MembersStateHandler.java -> synchronized ReplicaInformation [ ] getReplicasInformation ( )");
    assert stateMembers!=null;
    int stateReplicas = stateMembers.size();
    ReplicaInformation[] ret=new ReplicaInformation[stateReplicas];
    if (stateReplicas>0)
    {
      int index=0;
      Iterator iterator = stateMembers.entrySet().iterator();
      while (iterator.hasNext())
      {
        Map.Entry entry = (Map.Entry)(iterator.next());
        Integer replica = (Integer)(entry.getKey());
        MemberInfo info = (MemberInfo)(entry.getValue());
        ret[index++]=new ReplicaInformation(replica.intValue(),info.getCoordinatedMembers());
      }
      assert index==stateReplicas;
    }
    return ret;
  }


  //*************************************************************************************//
  //**************************** ASSIGN COORDINATOR *************************************//
  //*************************************************************************************//

  /**
   * Reassign the members to be coordinated. If there is no state members, no coordinator is assigned
   */
  void assignCoordinator(SortedSet memberSet)
  {
    Trace.code("MembersStateHandler.java -> void assignCoordinator ( SortedSet memberSet )");
    if (!memberSet.isEmpty())
    {
      int len = stateMembers.size();
      Integer onlyCoordinator = (len==1)? (Integer) (stateMembers.firstKey()) : null;
      Integer coordinator;
      Iterator members = memberSet.iterator();
      while(members.hasNext())
      {
        Integer groupMemberId = (Integer) (members.next());
        if (len==0)
          coordinator=null;
        else
        {
          if (onlyCoordinator!=null)
            coordinator = onlyCoordinator;
          else if (domainsUser.supportsCoordinatorElection())
            coordinator = requestCoordinator(groupMemberId);
          else
            coordinator=autoAssignCoordinator();

          MemberInfo newCoordinator = (MemberInfo) (stateMembers.get(coordinator));
          assert newCoordinator!=null;

          //the member is added to the memberInfo
          newCoordinator.addMemberToCoordinate(groupMemberId);
        }

        StatelessMemberInfo info = (StatelessMemberInfo) (statelessMembers.get(groupMemberId));
        assert info!=null;
        info.setCoordinator(coordinator);
      }
    }
  }

  //*************************************************************************************//
  //**************************** REQUEST COORDINATOR ************************************//
  //*************************************************************************************//

  /**
   * Requests a coordinator to the application
   */
  Integer requestCoordinator(Integer toCoordinate)
  {
    Trace.code("MembersStateHandler.java -> Integer requestCoordinator ( Integer toCoordinate )");
    return new Integer(domainsUser.getCoordinator(toCoordinate.intValue(), buildTransferStatusInfo()));
  }

  CoordinatorInformation[] buildTransferStatusInfo()
  {
    Trace.code("MembersStateHandler.java -> CoordinatorInformation [ ] buildTransferStatusInfo ( )");
    int len = stateMembers.size();
    CoordinatorInformation[] ret = new CoordinatorInformation[len];
    int idx = 0;
    Iterator it = stateMembers.entrySet().iterator();
    while(it.hasNext())
    {
      assert idx<len;
      Map.Entry entry = (Map.Entry)(it.next());
      int statusMember = ((Integer)(entry.getKey())).intValue();
      int[] currentCoordinations = ((MemberInfo)(entry.getValue())).getCoordinatedMembers();
      ret[idx++]=new CoordinatorInformation(statusMember, currentCoordinations);
    }
    assert idx==len;
    return ret;
  }

  //*************************************************************************************//
  //**************************** AUTO ASSIGN COORDINATOR ********************************//
  //*************************************************************************************//

  /**
   * locates a coordinator, current algorithm just returns the first state member
   */
  Integer autoAssignCoordinator()
  {
    Trace.code("MembersStateHandler.java -> Integer autoAssignCoordinator ( )");
    return (Integer) (stateMembers.firstKey());
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  boolean informationBuilt;

  SortedMap stateMembers; //sort by id! (Integer, MemberInfo)
  Map statelessMembers; //members to become statefull: Integer (GroupMemberId), StatelessMemberInfo
  List queuedStateMembers; //members that have achieved state, but it is not possible yet to use it
  //because informationBuilt is still false

  DomainsUser domainsUser;
  SortedSet membersToCoordinate; //created just to avoid creating the object repeatdly
}
