/**
  * File: MonitorImpl.java
  * Content: Monitor implementation. It is not directly the GroupMonitor implementation, but a class
  *       to be reused by any component with Monitor functionality
  * Author: LuisM Pena
  * Date: 03rd October 2000
  * Version: 0.51.00
  * Last change:
  *             6th November 2001: rewritten with state transfer support
  *             12th November 2001: state transfer support changes: monitor messages must
  *                   be received during state transfers, to fully support the notion of transactions
  *                   The state sent is the state when the view is installed
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.domains.MemberStateExceptionReason;
import sensei.middleware.domains.MonitorException;
import sensei.middleware.domains.MonitorMessage;
import sensei.middleware.domains.MonitorState;
import sensei.middleware.domains.State;

import sensei.util.Error;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
  * Monitor implementation. It is not directly the GroupMonitor implementation, but a class
  * to be reused by any component with Monitor functionality
  **/
public class MonitorImpl
{

//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  /**
   * Creates a monitor that must still receive the state.
   */
  public MonitorImpl(int subgroupId, DomainGroupHandler groupHandler)
  {
    Trace.code("MonitorImpl.java -> MonitorImpl ( int subgroupId , DomainGroupHandler groupHandler )");
    init(subgroupId, groupHandler);
    withState=false;
  }

  public void setSubgroupId(int subgroupId)
  {
    Trace.code("MonitorImpl.java -> void setSubgroupId ( int subgroupId )");
    factory.setSubgroup(subgroupId);
  }

  void init (int subgroupId, DomainGroupHandler groupHandler)
  {
    Trace.code("MonitorImpl.java -> void init ( int subgroupId , DomainGroupHandler groupHandler )");
    locks = new ArrayList();
    unprocessedMessages = new LinkedList();
    factory=new MonitorFactory(subgroupId, groupHandler);
    membersLockingAtViewChange=new int[0];
  }

  //*************************************************************************************
  //**************************** STATE OPERATIONS ***************************************
  //*************************************************************************************

  public State getState()
  {
    Trace.code("MonitorImpl.java -> State getState ( )");
    if (!withState)
      Trace.errorOnStateOperation(Trace.GET_STATE_OPERATION);

    return factory.createState(membersLockingAtViewChange);
  }

  public void setState(State s)
  {
    Trace.code("MonitorImpl.java -> void setState ( State s )");
    if (withState)
      Trace.errorOnStateOperation(Trace.SET_STATE_OPERATION);
    else
    {
      //first: store the state as it is received
      int[] stateLocks = ((MonitorState) s).locks;
      int nLocks = stateLocks.length;
      locks.clear();
      for(int i=0;i<nLocks;i++)
        locks.add(new Integer(stateLocks[i]));

      //second: update it with the new messages received from the installation of the view
      Iterator iterator = unprocessedMessages.iterator();
      while(iterator.hasNext())
      {
        UnprocessedMessage message=(UnprocessedMessage)iterator.next();
        if (message.lock)
          addLock(message.sender);
        else
          removeLock(message.sender);
      }
      unprocessedMessages.clear();
      withState = true;
    }
  }

  public void assumeState()
  {
    Trace.code("MonitorImpl.java -> void assumeState ( )");
    if (withState)
      Trace.errorOnStateOperation(Trace.ASSUME_STATE_OPERATION);
    else
    {
      locks.clear();
      unprocessedMessages.clear();
      withState=true;
    }
  }

  /**
   * Should be only called after being excluded!
   */
  public void freeResources()
  {
    Trace.code("MonitorImpl.java -> void freeResources ( )");
    assert !withState;

    if (locks!=null)
    {
      locks.clear();
      synchronized (this)
      {
        notifyAll();
        locks=null;
      }
      unprocessedMessages.clear();
      unprocessedMessages=null;
      factory.freeResources();
      factory=null;
      groupMemberId=null;
      membersLockingAtViewChange=null;
    }
  }

  //*************************************************************************************
  //**************************** LOCK OPERATIONS *****************************************
  //*************************************************************************************

  public void lock() throws MemberStateException
  {
    Trace.code("MonitorImpl.java -> void lock ( )");
    try
    {
      factory.castLock();
      waitToGetLock();
    }
    catch(MemberStateException igex){throw igex;}
    catch(NullPointerException rex){throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);}
    catch(Exception rex){Error.unhandledException(Consts.AREA,rex);}
  }

  //*************************************************************************************
  //**************************** UNLOCK *************************************************
  //*************************************************************************************

  /**
    * Unlocks the monitor
    **/
  public void unlock() throws MonitorException, MemberStateException
  {
    Trace.code("MonitorImpl.java -> void unlock ( )");
    if (haveITheLock())
    {
      try
      {
        factory.castUnlock();
      }
      catch(MemberStateException igex){throw igex;}
      catch(NullPointerException rex){throw new MemberStateException(MemberStateExceptionReason.MemberExcluded);}
      catch(Exception rex){Error.unhandledException(Consts.AREA,rex);}
    }
    else
      throw new MonitorException();
  }

  //*************************************************************************************
  //**************************** ADD LOCK ***********************************************
  //*************************************************************************************

  synchronized void addLock(Integer locking)
  {
    Trace.code("MonitorImpl.java -> synchronized void addLock ( Integer locking )");
    //Perhaps the lock is reentrant, in that case is just add on the beginning
    if (hasTheLock(locking))
      locks.add(0,locking);
    else
      locks.add(locking);
    notifyAll();
  }

  //*************************************************************************************
  //**************************** REMOVE LOCK ********************************************
  //*************************************************************************************

  synchronized void removeLock(Integer locking)
  {
    Trace.code("MonitorImpl.java -> synchronized void removeLock ( Integer locking )");
    if (!hasTheLock(locking))
      ;//Trace it!
    else
    {
      locks.remove(0);
      notifyAll();
    }
  }

  //*************************************************************************************
  //**************************** WAIT TO GET LOCK ***************************************
  //*************************************************************************************

  /**
    * Waits until the matching lock is received
    **/
  synchronized void waitToGetLock()
  {
    Trace.code("MonitorImpl.java -> synchronized void waitToGetLock ( )");
    while(withState && !haveITheLock())
      try{wait();}catch(InterruptedException ex){}
  }

  boolean haveITheLock()
  {
    Trace.code("MonitorImpl.java -> boolean haveITheLock ( )");
    return hasTheLock(groupMemberId);
  }

  synchronized boolean hasTheLock(Integer who)
  {
    Trace.code("MonitorImpl.java -> synchronized boolean hasTheLock ( Integer who )");
    return (locks!=null) && !locks.isEmpty() && locks.get(0).equals(who);
  }

  /***********************************************************************************************/
  /********************** METHODS ON THE APPLICATION GROUP MEMBER INTERFACE JAVA INTERFACE *******/
  /***********************************************************************************************/

  public void memberAccepted(int identity)
  {
    Trace.code("MonitorImpl.java -> void memberAccepted ( int identity )");
    groupMemberId=new Integer(identity);
    locks.clear();
  }

  public void excludedFromGroup()
  {
    Trace.code("MonitorImpl.java -> void excludedFromGroup ( )");
    //wake up every lock
    withState=false;
    locks.clear();
    synchronized (this){notifyAll();}
  }

  public void installView(View view)
  {
    Trace.code("MonitorImpl.java -> void installView ( View view )");
    if (withState)
    {
      removeInvalidLocks(view);
      storeCurrentState();
    }
    else
    {
      //clear received messages
      unprocessedMessages.clear();
    }
  }

  void storeCurrentState()
  {
    Trace.code("MonitorImpl.java -> void storeCurrentState ( )");
    int i=0, nLocks = locks.size();
    membersLockingAtViewChange = new int[nLocks];
    Iterator it = locks.iterator();
    while(it.hasNext())
    {
      assert i<nLocks;
      membersLockingAtViewChange[i++]=((Integer)it.next()).intValue();
    }
  }

  void removeInvalidLocks(View view)
  {
    Trace.code("MonitorImpl.java -> void removeInvalidLocks ( View view )");
    int invalidMembers[] = view.expulsedMembers;
    int length = invalidMembers.length;

    if (length> 0 && locks.size()>0)
    {
      boolean changed=false;
      synchronized(this)
      {
        Iterator iter = locks.iterator();
        while(iter.hasNext())
        {
          int owner = ((Integer) (iter.next())).intValue();
          for (int i=0; i < length ; i++)
            if (owner == invalidMembers[i])
            {
              changed = true;
              iter.remove();
              break;
            }
        }
        if (changed)
          notifyAll();
      }
    }
  }

  public void processCastMessage(int sender, Message message)
  {
    Trace.code("MonitorImpl.java -> void processCastMessage ( int sender , Message message )");
    try
    {
      if (message instanceof MonitorMessage)
      {
        Integer who=new Integer(sender);
        boolean lock=((MonitorMessage)message).lock;
        if (withState)
          if (lock)
            addLock(who);
          else
            removeLock(who);
        else //no state: store it
          unprocessedMessages.add(new UnprocessedMessage(who, lock));
      }
    }
    catch(Exception ex)
    {
      Error.unhandledException(Consts.AREA,ex);
    }
  }

  //*************************************************************************************//
  //**************************** INNER CLASS ********************************************//
  //*************************************************************************************//

  class UnprocessedMessage
  {
    public Integer sender;
    public boolean lock;
    public UnprocessedMessage(Integer sender, boolean lock){this.sender=sender; this.lock=lock;}
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  protected List unprocessedMessages; //[UnprocessedMessage]: messages received when the member has no state
  protected List locks; //[Integers]: current locks
  protected MonitorFactory factory; //to send messages middleware-independetly
  protected Integer groupMemberId;
  protected boolean withState;
  protected int membersLockingAtViewChange[]; //on each new view, it is needed to know the exact state!!
}
