/**
  * File: SubgroupsTasks.java
  * Content: Utility class defining the tasks to be performed on subgroups.
  *          It just defines and implements the tasks, but it must be subclasses
  *          in order to use those tasks
  *
  * Author: LuisM Pena
  * Date: 23rd April 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.View;
import sensei.middleware.domains.PhaseCoordination;
import sensei.middleware.domains.PhaseCoordinationHolder;
import sensei.middleware.domains.State;
import sensei.middleware.domains.SubgroupCoordination;

import sensei.util.IntWrapper;

import java.util.HashMap;
import java.util.Map;

/**
 * Utility class defining the tasks to be performed on subgroups.
 * It just defines and implements the tasks, but it must be subclasses
 * in order to use those tasks
 */
class SubgroupsTasks
{

  //*************************************************************************************

  interface Task
  {
    void call(STGroupMember subgroup) throws Exception;
  }

  interface StateTransferTask
  {
    PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
      throws Exception;
  }

  abstract class TaskWithPhase implements StateTransferTask
  {
    PhaseCoordinationHolder phaseHolder = new PhaseCoordinationHolder();
  }

  //*************************************************************************************

  class AssumeStateTask implements Task
  {
    //this task is only performed once!
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.assumeState();
    }
  };

  //*************************************************************************************

  class PostMemberAcceptanceTask implements Task
  {
    public PostMemberAcceptanceTask(int identity, GroupHandler handler, View view)
    {
      this.identity=identity;
      this.handler=handler;
      this.view=view;
    }
    public void set(View view)
    {
      this.view=view;
    }
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.memberAccepted(identity, handler, view);
    }
    int identity;
    GroupHandler handler;
    View view;
  };

  //*************************************************************************************

  class PostMemberExpulsionTask implements Task
  {
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.excludedFromGroup();
    }
  };

  //*************************************************************************************

  class InstallViewTask implements Task
  {
    public void set(View view){this.view=view;}
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.installView(view);
    }
    View view;
  };

  //*************************************************************************************

  class ChangeViewTask implements Task
  {
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.changingView();
    }
  };

  //*************************************************************************************

  class SendMessageTask implements Task
  {
    public void set(int sender, Message message){this.sender=sender;this.message=message;}
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.processPTPMessage(sender, message);
    }
    int sender;
    Message message;
  };

  //*************************************************************************************

  class CastMessageTask implements Task
  {
    public void set(int sender, Message message){this.sender=sender;this.message=message;}
    public void call(STGroupMember subgroup) throws Exception
    {
      subgroup.processCastMessage(sender, message);
    }
    int sender;
    Message message;
  };

  //*************************************************************************************

  class StartTransferTask extends TaskWithPhase
  {
    public void set(int target){targets[0] = target;}
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
       throws Exception
    {
      phaseHolder.value=coordination;
      subgroup.startTransfer(targets,phaseHolder);
      return phaseHolder.value;
    }
    int targets[]=new int[1];
  };

  //*************************************************************************************

  class StopTransferTask implements StateTransferTask
  {
    public void set(int joinings[], boolean transferFinished){this.joinings=joinings; finished=transferFinished;}
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
       throws Exception
    {
      subgroup.stopTransfer(joinings, finished);
      return coordination;
    }
    int joinings[];
    boolean finished;
  };

  //*************************************************************************************

  class SyncTransferTask extends TaskWithPhase
  {
    public void set(int coordinator){this.coordinator=coordinator;}
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
       throws Exception
    {
      phaseHolder.value=coordination;
      subgroup.syncTransfer(coordinator, phaseHolder);
      return phaseHolder.value;
    }
    int coordinator;
  };

  //*************************************************************************************

  class InterruptTransferTask extends TaskWithPhase
  {
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
       throws Exception
    {
      phaseHolder.value=coordination;
      subgroup.interruptTransfer(phaseHolder);
      return phaseHolder.value;
    }
  };

  //*************************************************************************************

  class ContinueTransferTask extends TaskWithPhase
  {
    public void set(int joinings[], Map joiningCoordinations){this.joinings=joinings; joiningCoords=joiningCoordinations;}
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
       throws Exception
    {
      SubgroupCoordination joining = (SubgroupCoordination)(joiningCoords.get(subgroupId));
      PhaseCoordination joiningPhase = (joining==null)? null : joining.phase;
      phaseHolder.value=coordination;
      subgroup.continueTransfer(joinings, phaseHolder, joiningPhase);
      return phaseHolder.value;
    }
    int joinings[];
    Map joiningCoords; //IntWrapper, SubgroupCoordination
  };

  //*************************************************************************************

  class GetStateTask extends TaskWithPhase
  {
    public void reset()
    {
      state = new HashMap();
    }
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
      throws Exception
    {
      phaseHolder.value=coordination;
      Object check=state.put(subgroupId, subgroup.getState(phaseHolder));
      assert check==null;
      return phaseHolder.value;
    }
    public Map getState()
    {
      return state;
    }
    Map state;
  };

  //*************************************************************************************

  class SendStateTask implements StateTransferTask
  {
    public void setState(Map states)
    {
      this.states=states;
    }
    public PhaseCoordination call(IntWrapper subgroupId, STGroupMember subgroup, PhaseCoordination coordination)
      throws Exception
    {
      //the transfer is only required if this subgroupId is the states map
      State state = (State)(states.get(subgroupId));
      if (state!=null)
        subgroup.setState(state, coordination);
      return coordination;
    }
    Map states;
  };

  //*************************************************************************************//
  //**************************** CONSTRUCTOR ********************************************//
  //*************************************************************************************//

  public SubgroupsTasks()
  {
    Trace.code("SubgroupsTasks.java -> SubgroupsTasks ( )");
    installViewTask = new InstallViewTask();
    changeViewTask = new ChangeViewTask();
    sendMessageTask = new SendMessageTask();
    castMessageTask= new CastMessageTask();
    postMemberExpulsionTask = new PostMemberExpulsionTask();

    startTransferTask = new StartTransferTask();
    stopTransferTask = new StopTransferTask();
    syncTransferTask = new SyncTransferTask();
    interruptTransferTask = new InterruptTransferTask();
    continueTransferTask = new ContinueTransferTask();
    getStateTask = new GetStateTask();
    sendStateTask = new SendStateTask();
  }

  //*************************************************************************************//
  //**************************** FREE RESOURCES *****************************************//
  //*************************************************************************************//
  public synchronized void freeResources()
  {
    Trace.code("SubgroupsTasks.java -> synchronized void freeResources ( )");
    postMemberAcceptanceTask=null;
    installViewTask=null;
    changeViewTask=null;
    sendMessageTask=null;
    castMessageTask=null;
    postMemberExpulsionTask=null;

    startTransferTask=null;
    stopTransferTask=null;
    syncTransferTask=null;
    interruptTransferTask=null;
    continueTransferTask=null;
    getStateTask=null;
    sendStateTask=null;
  }

  //*************************************************************************************//
  //**************************** DATA MEMBERS *******************************************//
  //*************************************************************************************//

  PostMemberAcceptanceTask postMemberAcceptanceTask; //this task is not created from the beginning!
  InstallViewTask installViewTask;
  ChangeViewTask changeViewTask;
  SendMessageTask sendMessageTask;
  CastMessageTask castMessageTask;
  PostMemberExpulsionTask postMemberExpulsionTask;

  StartTransferTask startTransferTask;
  StopTransferTask stopTransferTask;
  SyncTransferTask syncTransferTask;
  InterruptTransferTask interruptTransferTask;
  ContinueTransferTask continueTransferTask;
  GetStateTask getStateTask;
  SendStateTask sendStateTask;
}

