/**
  * File: CommunicationChannelsManager.java
  * Content: manages the communication channels, simplifying the
  *		ActiveGroupMember class. It doesn't define any tasks, that must be
  *   defined in subclasses.
  * Author: LuisM Pena
  * Date: 13th July 2000
  * Version: 0.97.00
  * Last change:
  *       v0.82. This is the main change for this version. Consensus is only reache
  *              when there is strict majority of numbers. The previous version allowed
  *              a parameter to decide when the consensus had to be strict or not. This
  *              supported cases like sending the token and detecting an error. In this
  *              case, the owner of the token could decide to keep the consensus even
  *              when not strict majority (and exactly half of the members) was alive.
  *              However, this overlooks the case when the token is received on the
  *              another side and then the communication is broken
  *       v0.95. InvalidateObject does not call freeResources, just does validManager=false,
  *              is the responsability of the owner to free the resourced
  *       v0.95. getFaultyMembers returns all the faulty members, including those that have
  *              never been included in a valid view. This behaviour is wrong. The function is
  *              renamed into getExcludedMembers()
  *       v0.95. Classes changed into interfaces (Vector-->List, etc)
  *       v0.95. Allowed weak consensus, upon a configuration value
  *
  * Comments:
  *				The purpose of this class is to simplify the ActiveGroupMember tasks.
  *				By itself, this class should make synchronized most of the public methods,
  *				but this would likely block the system, if different tasks int the
  *				ActiveGroupMember would try to access the channels Manager. For example,
  *				a task sending the view and another one receiving it. For this reason,
  *				the synchronization is considered to be at ActiveGroupMember level
  *
  * 			Currently, the own identity is obtained with _is_equivalent, which is
  *				not existing in RMI (easy to solve) but is not reliable in CORBA (there
  *				is no security about the answer). It should be changed to calls to the
  *				members to obtain their identity. This would be done only once, therefore
  *				is not a) a performance problem and b) very contradictory with the exclusive
  *				access to other member through the channels.
  **/

package sensei.gms;

import sensei.middleware.gms.SenseiGMSMember;
import sensei.middleware.gms.InternalView;

import sensei.middleware.util.ObjectsHandling;

import sensei.util.Error;
import sensei.util.IntWrapper;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

/**
  * Manages the communication channels belonging to one member, simplifying the
  *		ActiveGroupMember class. It doesn't define any tasks, that must be
  *   defined in subclasses.
  **/
class CommunicationChannelsManager
{
  /**
    * The manager is built defining from first moment the channels
    * to be included in the ActiveGroupMember
    * @param thisSenseiGMSMember the SenseiGMSMember owning this channel
    * @param initialView View, containing the members for which is needed to create channels
    **/
  public CommunicationChannelsManager(SenseiGMSMember thisSenseiGMSMember, InternalView initialView)
  {
    Trace.code("CommunicationChannelsManager.java -> CommunicationChannelsManager ( SenseiGMSMember thisSenseiGMSMember , InternalView initialView )");
    assert thisSenseiGMSMember!=null && initialView!=null;
    assert initialView.members.length == initialView.memberIds.length;

    threadGroup							= new ThreadGroup(newName());
    channels 								= new Vector();
    rank 										= -1;
    nextChannel							= -1;
    consensusCondition      =  0;
    cmIntWrapper						= new IntWrapper();
    vmIntWrapper						= new IntWrapper();
    vm2IntWrapper						= new IntWrapper();
    thisMember							= obtainMyIdentity(thisSenseiGMSMember, initialView);
    addedMembers						= new HashSet();
    faultyMembers						= new HashSet();
    validMembers						= new HashSet();
    excludedMembers         = new ArrayList();
    validManager						= true;
    newFaultyMembers				= false;
    activeCommunications    = true;
    autoExclusions          = 0;

    createInitialCommunicationChannels(initialView);

    assert rank != -1;
  }

//*************************************************************************************//
//**************************** STOP COMMUNICATIONS ************************************//
//*************************************************************************************//

  /**
    * Stops the communications, no messages or views will be sent afterwards
    **/
  public void stopCommunications()
  {
    Trace.code("CommunicationChannelsManager.java -> void stopCommunications ( )");
    activeCommunications = false;
  }

//*************************************************************************************//
//**************************** CONSOLIDATE VIEW ***************************************//
//*************************************************************************************//

  /**
    * Sets the current members as the consolidated ones in a installed view
    **/
  public void consolidateView()
  {
    Trace.code("CommunicationChannelsManager.java -> void consolidateView ( )");
    if (isValidObject())
    {
      assert !newFaultyMembers;

      addedMembers.clear();
      faultyMembers.clear();
      excludedMembers.clear();

      consensusCondition = (channels.size() - weakConsensus)/ 2;
      autoExclusions=0;
    }
  }

//*************************************************************************************//
//**************************** FAULTY MEMBERS TO COMMUNICATE **************************//
//*************************************************************************************//

  /**
    * Returns true if the manager has any faulty channel that must be communicated to
    * the other members.
    **/
  public boolean faultyMembersToCommunicate()
  {
    Trace.code("CommunicationChannelsManager.java -> boolean faultyMembersToCommunicate ( )");
    return isValidObject() && newFaultyMembers;
  }

//*************************************************************************************//
//**************************** CONSENSUS **********************************************//
//*************************************************************************************//

  /**
    * Returns true if there is consensus, that is, if at least half of the members in
    * the previous view are still alive now
    * @returns true if there is consensus
    **/
  public boolean consensus()
  {
    Trace.code("CommunicationChannelsManager.java -> boolean consensus ( )");
    return isValidObject() && ((channels.size() + autoExclusions - addedMembers.size()) > consensusCondition);
  }

//*************************************************************************************//
//**************************** GET VIEW ***********************************************//
//*************************************************************************************//

  /**
    * Obtains the list of members from the communication channels. Only the alive members
    * are returned. After this method, a call to faultyMembersToCommunicate will return false
    * @param holder the InternalView to contain the returned value, intended to minimize
    *    the number of object creations
    * @param includeAddedMembers set to true if the view returned must contains as well
    *    the information on addedMembers
    **/
  public InternalView getView(InternalView holder, boolean includeAddedMembers)
  {
    Trace.code("CommunicationChannelsManager.java -> InternalView getView ( InternalView holder , boolean includeAddedMembers )");
    assert holder!=null;

    if (isValidObject())
    {
      int length = channels.size();
      holder.members = new SenseiGMSMember[length];
      holder.memberIds = new int[length];
      while(length-->0)
      {
        CommunicationChannel channel = (CommunicationChannel) channels.get(length);
        holder.members[length] = channel.getLink();
        holder.memberIds[length] = channel.getLinkId();
      }
      newFaultyMembers = false;
      if (includeAddedMembers)
      {
        length = addedMembers.size();
        holder.newMembers = new int[length];
        Iterator iterator = addedMembers.iterator();
        while(iterator.hasNext())
          holder.newMembers[--length]=((IntWrapper)(iterator.next())).i;
      }
      else
        holder.newMembers = new int[0];
    }
    return holder;
  }

//*************************************************************************************//
//**************************** GET AUTO EXCLUSION VIEW ********************************//
//*************************************************************************************//

  /**
    * This call is equivalent to getView with includeAddedMembers=true, and removing the
    * own member identity. If this is the only member, it returns false.
    * @param includeAddedMembers set to true if the view returned must contains as well
    *    the information on addedMembers
    **/
  public boolean getAutoExclusionView(InternalView holder)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean getAutoExclusionView ( InternalView holder )");
    assert isValidMember(thisMember);

    boolean ret=false;

    if (isValidObject())
    {
      InternalView view = getView(holder, true);
      int length = view.memberIds.length;
      if (length > 1)
      {
        ret=true;
        SenseiGMSMember[] finalList = new SenseiGMSMember[length-1];
        int[] finalIds = new int[length-1];
        for (int i = --length ; i>=0 ; --i)
          if (holder.memberIds[i]!=thisMember)
          {
            assert length>0;
            finalIds[--length]=holder.memberIds[i];
            finalList[length]=holder.members[i];
          }
        holder.memberIds=finalIds;
        holder.members=finalList;
        assert length==0;
        if (!isValidInstalledMember(thisMember))
        {
          length = holder.newMembers.length;
          int finalAddedMembers[]=new int[--length];
          for (int i=length;i>=0;--i)
            if (holder.newMembers[i]!=thisMember)
            {
              assert length>0;
              finalAddedMembers[--length]=holder.newMembers[i];
            }
          holder.newMembers=finalAddedMembers;
          assert length==0;
        }
      }
    }
    return ret;
  }

//*************************************************************************************
//**************************** GET EXCLUDED MEMBERS ***********************************
//*************************************************************************************

  /**
    * Returns the list of faulty members
    **/
  public int[] getExcludedMembers()
  {
    Trace.code("CommunicationChannelsManager.java -> int [ ] getExcludedMembers ( )");
    int[] ret;
    if (isValidObject())
    {
      int len = excludedMembers.size();
      ret = new int[len];
      Iterator iterator = excludedMembers.iterator();
      while(iterator.hasNext())
        ret[--len]=((IntWrapper)(iterator.next())).i;
    }
    else
      ret=new int[0];
    return ret;
  }

//*************************************************************************************//
//**************************** CHANGE MEMBERS *****************************************//
//*************************************************************************************//

  /**
    * Changes the members associated: new channels are created, and those
    * linking to members removed are considered as faulty
    * @param checkAll is true if the view contains valid addedMembers to be checked
	  *                                                                      	and the faulty members must be already checked
    * @returns true if the action means an autoexclusion from the viewSender
    **/
  public synchronized boolean changeMembers (int viewSender, InternalView view, boolean checkAll)
  {
    Trace.code("CommunicationChannelsManager.java -> synchronized boolean changeMembers ( int viewSender , InternalView view , boolean checkAll )");
    //some work here!
    //Iterate over the member parameter and over the communication channels.
    //-If the member is in the list of faulty members, this member must publish a view
    //  showing that member as faulty: mark it in the list of faulty members as not
    //  known to be faulty.
    //-If the member is in the list of channels, it's a valid already known member, but
    //  all the channels in the middle since the last recogniced and this one must be
    //  considered faulty and included in the list of faulty members as known to be faulty.
    //-Otherwise, is a new member, insert it as new member. It is checked with the internalView
    //  received that is has been marked as new member
    assert (view!=null) && (view.members.length == view.memberIds.length);

    boolean ret=false;

    if (isValidObject())
    {
      SenseiGMSMember members[] = view.members;
      int memberIds[] = view.memberIds;
      int lenMembers  = memberIds.length;
      int channel     = 0;

      boolean probablyAutoExclusion = !isValidMember(viewSender);

      for (int member = 0 ; member < lenMembers ; member ++)
      {
        int nextMember = memberIds[member];
        cmIntWrapper.i = nextMember;
        if (faultyMembers.contains(cmIntWrapper))
          assert !checkAll;
        else
        {
          int rankNextMember = getRank(nextMember);
          if (rankNextMember==-1)	//new member
            addNewChannel(members[member], nextMember, channel);
          else if (channel != rankNextMember)
            removeChannels(channel,rankNextMember);
          channel++;
        }
      }
      removeChannels(channel,channels.size());

      if (!isValidMember(viewSender))
      {
        if (probablyAutoExclusion)
        {
          assert autoExclusions>0;
        }
        else
        {
          ++autoExclusions;
          ret=true;
        }
      }

      //it is only checked the number of added members
      assert !checkAll || view.newMembers.length==addedMembers.size();
    }
    return ret;
  }

//*************************************************************************************//
//**************************** GET MY ID **********************************************//
//*************************************************************************************//

  /**
    * Returns the identity of this member
    **/
  public int getMyId()
  {
    Trace.code("CommunicationChannelsManager.java -> int getMyId ( )");
    return thisMember;
  }

//*************************************************************************************//
//**************************** IS PREV MEMBER IN VIEW *********************************//
//*************************************************************************************//

  /**
    * Returns true if the channel specified is the precedent to the owner in the view
    **/
  public boolean isPrevMemberInView(int member)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean isPrevMemberInView ( int member )");
    int itsRank		=	getRank(member);
    int previous	= rank==0? channels.size()-1 : rank-1;
    return itsRank == previous;
  }

//*************************************************************************************//
//**************************** IS VALID MEMBER ****************************************//
//*************************************************************************************//

  public boolean isValidMember(SenseiGMSMember member)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean isValidMember ( SenseiGMSMember member )");
    if (validManager)
    {
      Iterator it = channels.iterator();
      while (it.hasNext())
        if (ObjectsHandling.areEquivalent(member,((CommunicationChannel)(it.next())).getLink()))
          return activeCommunications;
    }
    return false;
  }

//*************************************************************************************//
//**************************** IS VALID MEMBER ****************************************//
//*************************************************************************************//

  public boolean isValidMember(int member)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean isValidMember ( int member )");
    vmIntWrapper.i=member;
    return activeCommunications && validMembers.contains(vmIntWrapper);
  }

//*************************************************************************************//
//**************************** IS VALID MEMBER ****************************************//
//*************************************************************************************//

  public boolean isValidInstalledMember(int member)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean isValidInstalledMember ( int member )");
    vm2IntWrapper.i=member;
    return activeCommunications && validMembers.contains(vm2IntWrapper) && !addedMembers.contains(vm2IntWrapper);
  }

//*************************************************************************************//
//**************************** MAKE REQUEST TO EVERY CHANNEL **************************//
//*************************************************************************************//

  /**
    * Makes asynchronous request to the channels, in two steps.
    * First, they're contacted to make the asynchronous communication and in a second
    * step the channel waits for the answer. It returns true if the request is
    * successfull (it depends mainly on the ChannelRequester). Those channels giving
    * errors are automatically removed.
    * @param requester A ChannelRequester member who does the individual requests
    * @returns true if every contacted member is alive and accepts the request
    **/
  boolean makeRequestToEveryChannel(ChannelRequester requester)
  {
    Trace.code("CommunicationChannelsManager.java -> boolean makeRequestToEveryChannel ( ChannelRequester requester )");
    boolean sentToAll = isValidObject();

    //two steps.
    //first: do the request asynchronously
    //second: wait for the answer
    for (int step=1;step<=2;step++)
    {

      boolean sentToThis=false;
      CommunicationChannel channel = null;
      int member = 0;

      while(isValidObject() && (member < channels.size()))
        try
        {
          channel = (CommunicationChannel) channels.get(member);
          if (step==1)
            sentToThis = channel.invoke(requester);
          else
            sentToThis = channel.waitChannel();
        }
        catch(Exception ex)
        {
          sentToThis = false;
          Trace.wrongCommunication(channel.getLinkId(), step, ex.toString());
        }
        finally
        {
          if (sentToThis)
            member++;
          else
          {
            Trace.wrongCommunication(channel.getLinkId(), step, requester.getClass().getName().substring(requester.getClass().getPackage().getName().length()+1));

            sentToAll = false;
            //if the own member has failed (for example, taking too long time), try
            //to finish in a right way
            if (member == rank)
              invalidateObject();
            else
              faultyMember(member);
          }
        }
    }

    return sentToAll;
  }

//*************************************************************************************//
//**************************** FREE RESOURCES *****************************************//
//*************************************************************************************//

  /**
    * Releases any resource associated to this object (~destructor)
    **/
  public void freeResources()
  {
    Trace.code("CommunicationChannelsManager.java -> void freeResources ( )");
    if (channels!=null)
    {
      validManager = false;

      freeChannels();

      threadGroup = null;
      channels 		= null;

      addedMembers.clear();
      faultyMembers.clear();
      excludedMembers.clear();
      validMembers.clear();

      addedMembers = null;
      faultyMembers= null;
      excludedMembers=null;
      validMembers=null;

      vmIntWrapper = null;
      vm2IntWrapper = null;
      cmIntWrapper = null;

      Trace.releaseObject("CommunicationChannelsManager");
    }
  }

//*************************************************************************************//
//**************************** FREE CHANNELS ******************************************//
//*************************************************************************************//

  /**
    * Releases the channels and free their resources
    **/
  void freeChannels()
  {
    Trace.code("CommunicationChannelsManager.java -> void freeChannels ( )");
    Iterator it = channels.iterator();
    while (it.hasNext())
    {
      ((CommunicationChannel) it.next()).freeResources();
    }
    channels.clear();
  }

//*************************************************************************************//
//**************************** CREATE INITIAL COMMUNICATION CHANNELS ******************//
//*************************************************************************************//

  /**
    * Creates the communication channels, init channels & nextChannel
    **/
  void createInitialCommunicationChannels(InternalView initialView)
  {
    Trace.code("CommunicationChannelsManager.java -> void createInitialCommunicationChannels ( InternalView initialView )");
    //assert to verify that there are no channels created
    assert channels.size()==0;

    //create the channels
    for ( int i=0 ; i < initialView.members.length ; i++ )
    {
      channels.add(new CommunicationChannel(initialView.members[i], initialView.memberIds[i], threadGroup));
      validMembers.add(new IntWrapper(initialView.memberIds[i]));
    }

    //create the initial added Channels
    for (int i=0;i <initialView.newMembers.length;i++)
      addedMembers.add(new IntWrapper(initialView.newMembers[i]));

    //fix view variables
    setVariablesForView();
  }

//*************************************************************************************//
//**************************** SET VARIABLES FOR VIEW *********************************//
//*************************************************************************************//

  /**
    * When a view is changed, some variables, like the rank or the
    * next members must be set.
    **/
  void setVariablesForView()
  {
    Trace.code("CommunicationChannelsManager.java -> void setVariablesForView ( )");
    //obtain the rank
    rank=getRank(thisMember);
    assert rank!=-1;

    nextChannel = findNextChannel(rank);
  }

//*************************************************************************************//
//**************************** GET RANK ***********************************************//
//*************************************************************************************//

  /**
    * Finds the channel associated to the member specified, returning
    * -1 if not found or its position in the channels vector otherwise
    **/
  int getRank(int member)
  {
    Trace.code("CommunicationChannelsManager.java -> int getRank ( int member )");
    int lenMembers = channels.size();
    for (int rank = 0 ; rank < lenMembers ; rank++ )
    {
      CommunicationChannel cc=(CommunicationChannel)channels.get(rank);
      if (cc.getLinkId()==member)
        return rank;
    }
    return -1;
  }

//*************************************************************************************//
//**************************** FIND NEXT CHANNEL **************************************//
//*************************************************************************************//

  /**
    * Finds the channel after the owner (the next element in the view)
    **/
  int findNextChannel(int after)
  {
    Trace.code("CommunicationChannelsManager.java -> int findNextChannel ( int after )");
    int channel=after+1;
    if (channel>=channels.size())
      channel=0;
    return channel;
  }

//*************************************************************************************//
//**************************** ADD NEW CHANNEL ****************************************//
//*************************************************************************************//

  /**
    * Creates a new communication channel for the new member in the position specified
    * The channel is inserted in the channels Vector and the member in the AddedMembers vector
    * Note that serVariablesForView should be called after it!
    * @param iterator an iterator over the GroupMembers to create
    **/
  void addNewChannel(SenseiGMSMember member, int memberId, int position)
  {
    Trace.code("CommunicationChannelsManager.java -> void addNewChannel ( SenseiGMSMember member , int memberId , int position )");
    channels.add(position, new CommunicationChannel(member, memberId, threadGroup));
    IntWrapper wrapped = new IntWrapper(memberId);
    validMembers.add(wrapped);
    addedMembers.add(wrapped);

    setVariablesForView();
  }

//*************************************************************************************//
//**************************** REMOVE CHANNELS ****************************************//
//*************************************************************************************//

  /**
    * Remove the channels specified from the channels Vector,
    * from the first till the last -1, They're signed as not needed to be communicated
    * inserting them in the faultyMembers list as known to be
    * faulty. They are as well removed from the addedMembers vector
    * Modifies: channels, faultyMembers, addedMembers
    * Channels are resource-freed.
    **/
  void removeChannels(int first, int last)
  {
    Trace.code("CommunicationChannelsManager.java -> void removeChannels ( int first , int last )");
    assert last>=first;

    //remove intermediate members
    while(first<=--last)
      removeChannel(last);
  }

//*************************************************************************************//
//**************************** REMOVE CHANNEL *****************************************//
//*************************************************************************************//

  /**
    * Remove the channel specified from the channels Vector,
    * inserting them in the faultyMembers list as known to be
    * faulty. They are as well removed from the addedMembers vector
    * Modifies: channels, faultyMembers, addedMembers
    * Channels are resource-freed.
    **/
  void removeChannel(int channel)
  {
    Trace.code("CommunicationChannelsManager.java -> void removeChannel ( int channel )");
    assert channel != rank;
    assert channel < channels.size();
    assert channel >= 0;

    CommunicationChannel removedChannel = (CommunicationChannel)channels.remove(channel);
    IntWrapper removedMember = new IntWrapper (removedChannel.getLinkId());
    if (!addedMembers.remove(removedMember))
      excludedMembers.add(removedMember);
    validMembers.remove(removedMember);
    faultyMembers.add(removedMember);
    removedChannel.freeResources();

    setVariablesForView();	//changes probably the variable rank
  }

//*************************************************************************************//
//**************************** FAULTY MEMBER ******************************************//
//*************************************************************************************//

  /**
    * Removes the channel and mark the view as having faulty members
    **/
  void faultyMember(int channel)
  {
    Trace.code("CommunicationChannelsManager.java -> void faultyMember ( int channel )");
    removeChannel(channel);
    newFaultyMembers = true;
  }

//*************************************************************************************//
//**************************** IS VALID OBJECT ****************************************//
//*************************************************************************************//

  /**
    * Verifies if the current object is valid
    **/
  public boolean isValidObject()
  {
    Trace.code("CommunicationChannelsManager.java -> boolean isValidObject ( )");
    return validManager && activeCommunications;
  }

//*************************************************************************************//
//**************************** INVALIDATE OBJECT **************************************//
//*************************************************************************************//

  /**
    * Invalidates this object, should not be used anymore
    **/
  void invalidateObject()
  {
    Trace.code("CommunicationChannelsManager.java -> void invalidateObject ( )");
    validManager=false;
  }

//*************************************************************************************//
//**************************** OBTAIN MY IDENTITY **************************************//
//*************************************************************************************//

  /**
    * Obtains the identity of this member in the view.
    * An assertion is raised if it is not found
    * Returns INVALID_IDENTITY if not found
    **/
  int obtainMyIdentity(SenseiGMSMember member, InternalView view)
  {
    Trace.code("CommunicationChannelsManager.java -> int obtainMyIdentity ( SenseiGMSMember member , InternalView view )");
    SenseiGMSMember members[] = view.members;
    int len = members.length;
    while (len-->0)
      if (ObjectsHandling.areEquivalent(member, members[len]))
        return view.memberIds[len];
    Error.error(Consts.GMS, "CommunicationChannelsManager:obtainMyIdentity:not found");
    return Consts.INVALID_MEMBER_ID;	//never
  }

//*************************************************************************************//
//**************************** SEND VIEW IF NEEDED ************************************//
//*************************************************************************************//

  /**
    * Allocates a new name for this object, garantized to be unique
    **/
  static private String newName()
  {
    Trace.code("CommunicationChannelsManager.java -> String newName ( )");
    return "CommunicationCenter." + Long.toString(++centerNumber);
  }


//*************************************************************************************//
//**************************** DATA MEMBERS *******************************************//
//*************************************************************************************//

  int thisMember;
  ThreadGroup threadGroup;

  boolean validManager, activeCommunications;

  int rank;													//the position of this member in the view
  int nextChannel;									//next channel in the view
  int consensusCondition;      		  //valid members must be > than the consensus condition
  int autoExclusions;               //number of autoexclusions in the current view

  List channels;									  //communicationChannels
  Set validMembers;							    //communicationChannels, to speed up consults (GroupMemberIds)
  Set addedMembers;							    //members being added (GroupMemberIds)
  Set faultyMembers;						    //members faulty  (GroupMemberIds)
  List excludedMembers;             //members excluded on this view (GroupMemberIds)
  boolean newFaultyMembers;					//set to true if the handler has found new errors

  IntWrapper cmIntWrapper, vmIntWrapper, vm2IntWrapper;
  static int centerNumber=0;				//used to name the threadgroups
  static int weakConsensus=Consts.WEAK_CONSENSUS? 1 : 0;
}
