/**
  * File: TokenRecoverer.java
  * Content: Defines the way to recover a token. This class, after some time
  *          elapsed without any communication, will contact the rest of
  *          members in the group to verify their liveness and which member
  *          must recover the token.
  * Author: LuisM Pena
  * Date: 17th August 2000
  * Version: 0.97.00
  * Last change:
  * Comments: freeResources && object vailidy complete
  **/

package sensei.gms;

import sensei.middleware.gms.InternalView;
import sensei.middleware.gms.InternalViewId;
import sensei.middleware.gms.MessageId;
import sensei.middleware.gms.SenseiGMSMember;

import sensei.util.Timer;
import sensei.util.TimerListener;

/**
  * TokenRecoverer is the part of the SenseiGMSMember that handles the token recovery.
  * After some time elapsed without any communication,
  * will contact the rest of members in the group to verify their liveness and
  * which member must recover the token.<p>
  * When this class considers the token lost, will call to the associated
  * ActivatedGroupMember to do the real recovery; this must check with the
  * TokenRecoverer to verify the recovery: this is needed to avoid race conditions
  * when a communication is received when the token is going to be recovered, for
  * example.
  **/
class TokenRecoverer implements TimerListener
{

//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  /**
    * It's needed to know the SenseiGMSMember handled and the ActiveGroupMember that
    * creates this recoverer (will be alerted if the token get lost).
    * This object will start immediately to control other members, therefore it must
    * not be created until the member becomes part of a group
    * @param activeGroupMember The creator if this object
    * @param thisMember The SenseiGMSMember that this ActiveGroupMember represent
    * @param view The view to handle
    **/
  public TokenRecoverer(ActiveGroupMember activeGroupMember, SenseiGMSMember thisMember, InternalView view)
  {
    Trace.code("TokenRecoverer.java -> TokenRecoverer ( ActiveGroupMember activeGroupMember , SenseiGMSMember thisMember , InternalView view )");
    this.thisMember					= thisMember;
    this.activeGroupMember 	= activeGroupMember;
    this.view            		= view;

    channelsManager					= null;
    messageIdToConfirm			=	null;
    stopped									= false;
    acceptCommunications		= true;
    timer										= new Timer(this, RECOVER_TOKEN_PERIOD,0L);
    timer.start();
  }

//*************************************************************************************//
//**************************** FREE RESOURCES *****************************************//
//*************************************************************************************//

  /**
    * Releases any resource associated to this object (~destructor)
    **/
  public synchronized void freeResources()
  {
    Trace.code("TokenRecoverer.java -> synchronized void freeResources ( )");
    if (isValidObject())
    {
      timer.release();

      if (channelsManager!=null)
        channelsManager.freeResources();

      activeGroupMember	=	null;
      view							= null;
      channelsManager		= null;
      timer							= null;
      messageIdToConfirm= null;

      Trace.releaseObject("TokenRecoverer");
    }
  }

//*************************************************************************************//
//**************************** VERIFY TOKEN IS LOST ***********************************//
//*************************************************************************************//

  /**
    * To avoid race conditions, this method must be called once the ActiveGroupMember
    * is activated to verify that no communications have happened on the meanwhile.
    * This operation FROZES the tokenRecoverer, which will pay no attention to new
    * communications until the operation tokenIsLostVerified is called. This allows to
    * the caller to obtain the members, messageId, etc without external problems.
    * Note that is only frozen if the return value is True
    **/
  public synchronized boolean verifyTokenIsLost(CommunicationChannelsManager ccm)
  {
    Trace.code("TokenRecoverer.java -> synchronized boolean verifyTokenIsLost ( CommunicationChannelsManager ccm )");
    boolean ret = ccm == channelsManager && ccm != null;
    if (ret)
      acceptCommunications = false;	//frozen!
    return ret;
  }

//*************************************************************************************//
//**************************** TOKEN IS LOST VERIFIED *********************************//
//*************************************************************************************//

  /**
    * Allows the tokenRecoverer to answer to communications received
    * See verifyTokenIsLost description.
    **/
  public synchronized void tokenIsLostVerified()
  {
    Trace.code("TokenRecoverer.java -> synchronized void tokenIsLostVerified ( )");
    acceptCommunications = true;
  }

//*************************************************************************************//
//**************************** PROCESS LAST MESSAGE ***********************************//
//*************************************************************************************//

  /**
    * After having activated the ActiveGroupMember, and once this has checked that
    * the lost is indeed lost, this method must be called to verify if the last message
    * had to be processed or not. The result is undefined if the correct method call
    * order is not kept.
    * @return null if there is no need of any confirmation, otherwise, it's returned
    *              the identity of the message to confirm
    **/
  public synchronized MessageId getMessageToConfirm()
  {
    Trace.code("TokenRecoverer.java -> synchronized MessageId getMessageToConfirm ( )");
    assert isValidObject();
    return messageIdToConfirm;
  }

//*************************************************************************************//
//**************************** STOP ***************************************************//
//*************************************************************************************//

  /**
    * Stops the TokenRecovery activity; it can be resumed with resume()
    **/
  public synchronized void stop()
  {
    Trace.code("TokenRecoverer.java -> synchronized void stop ( )");
    assert isValidObject();
    stopped = true;
    channelsManager = null;
    timer.stop();
  }

//*************************************************************************************//
//**************************** RESUME *************************************************//
//*************************************************************************************//

  /**
    * Resumes a stopped TokenRecoverer
    **/
  public synchronized void resume()
  {
    Trace.code("TokenRecoverer.java -> synchronized void resume ( )");
    assert isValidObject();
    if (stopped)
    {
      stopped = false;
      timer.restart();
    }
  }

//*************************************************************************************//
//**************************** COMMUNICATION RECEIVED *********************************//
//*************************************************************************************//

  /**
    * Communicates to this object that there have been a valid communication; the timer
    * will be restarted.
    * @return false if the tokenRecoverer is frozen, what means that the SenseiGMSMember
    *    is recovering the token, therefore no communications should be received!
    **/
  public synchronized boolean communicationReceived()
  {
    Trace.code("TokenRecoverer.java -> synchronized boolean communicationReceived ( )");
    assert isValidObject();
    if (acceptCommunications)
    {
      channelsManager = null;
      if (!stopped)
        timer.restart();
    }
    return acceptCommunications;
  }

//*************************************************************************************//
//**************************** COMMUNICATION RECEIVED *********************************//
//*************************************************************************************//

  /**
    * Communicates to this object that there have been a valid communication; the timer
    * will be restarted.
    * @return false if the tokenRecoverer is frozen, what means that the SenseiGMSMember
    *    is recovering the token, therefore no communications should be received!
    **/
  public synchronized boolean communicationReceived(InternalView newView)
  {
    Trace.code("TokenRecoverer.java -> synchronized boolean communicationReceived ( InternalView newView )");
    if (acceptCommunications)
    {
      channelsManager = null;
      view = newView;
      communicationReceived();
    }
    return acceptCommunications;
  }

//*************************************************************************************//
//**************************** TIMER EXPIRED ******************************************//
//*************************************************************************************//

  /**
    * Method called if the timer to recover the token expires
    * @returns true if the timer can continue its execution
    **/
  public boolean timerExpired()
  {
    Trace.code("TokenRecoverer.java -> boolean timerExpired ( )");
    boolean ret = true;

    Trace.tokenRecovery(Trace.TokenRecoveryAsking);

    //only problem possible: timerExpired is recalled when acceptCommunications is false
    //in this case, everything should be frozen!!
    // case a - a communication is received before the synchronized code in this method
    //          Timer will be restarted, and channelsManager is not changed. The timer
    //          is stopped manually in the synchronized code to avoid it.
    // case b - a communication is received after the synchronized code is executed.
    //          In that case the channelsManager is set to null and acceptCommunications
    //           would never become false.

    TokenRecovererCommunicationChannelsManager ccm;
    InternalViewId viewId;
    synchronized (this)
    {
      ccm							= new	TokenRecovererCommunicationChannelsManager(thisMember, view);
      viewId 					= new InternalViewId (view.viewId.id, view.viewId.installing);
      channelsManager = ccm;
      timer.stop();		//avoid any restart!. If after this is restarted, channelsManager
                      //will be set to null, but if it was restarted just between the call
                      //to timerExpired() and this synchronized code, the timer would be
                      //restarted and the channelsManager unset to null.
    }
    OutBoolean processLastMessage	= new OutBoolean();
    messageIdToConfirm						=	new MessageId(0,0);

    if (ccm.sendRequestToGenerateToken(viewId, messageIdToConfirm, processLastMessage)
        && ccm==channelsManager)	//the channelsManager could have changed on the meanwhile!
    {
      //yes, token can be recovered (there is nobody that has replied NOT, although they could
      //have answered all of them that this member is not in the group!
      if (!processLastMessage.value)
        messageIdToConfirm = null;

      //The activeGroupMember is given notice of it
      ret = activeGroupMember.tokenLost(ccm);

    }
    else
    {
      Trace.tokenRecovery(Trace.TokenRecoveryDropped);
      timer.start();	//start the timer to try again later
    }

    return ret;
  }


//*************************************************************************************//
//**************************** IS VALID OBJECT ****************************************//
//*************************************************************************************//

  /**
    * Verifies that the object is valid (not yet released)
    **/
  boolean isValidObject()
  {
    Trace.code("TokenRecoverer.java -> boolean isValidObject ( )");
    return timer!=null;
  }

//*************************************************************************************//
//**************************** DATA MEMBERS *******************************************//
//*************************************************************************************//

  SenseiGMSMember thisMember;														//member owning this object
  ActiveGroupMember activeGroupMember;					//ActiveGroupMember to alert if token get lost
  InternalView view;																		//view to verify

  MessageId messageIdToConfirm;									//if the token is lost, this member contain
                                                //the message to confirm, or null if there is none
  boolean stopped;
  boolean acceptCommunications;

  TokenRecovererCommunicationChannelsManager channelsManager;	//manager of channels

  Timer timer;																	//timer to check token loose.
  static final long RECOVER_TOKEN_PERIOD = Consts.RECOVER_TOKEN_PERIOD;
}
