/**
  * File: TokenState.java
  * Content: Simple class keeping the state of the token
  * Author: LuisM Pena
  * Date: 25th November 2000
  * Version: 0.97.00
  * Last change:
  * Comments: only the access to change the tokenStatus needs to be synchronized
  *
  **/

package sensei.gms;

import sensei.middleware.gms.Token;


/**
  * Simple class keeping the state of the token
  **/
class TokenState
{

//*************************************************************************************//
//**************************** CONSTRUCTOR ********************************************//
//*************************************************************************************//

  /**
    * The constructor requires the channel manager to use to send the token
    * The initialToken can be null, In that case, one is just generated
    **/
  public TokenState(int thisMemberId, ViewHandler channelsManager, Token initialToken)
  {
    Trace.code("TokenState.java -> TokenState ( int thisMemberId , ViewHandler channelsManager , Token initialToken )");
    assert channelsManager!=null;
    this.channelsManager=channelsManager;
    lastGeneratedToken = new Token(thisMemberId, 0);
    if (initialToken==null)
      token = lastGeneratedToken;
    else
      token=initialToken;
    tokenStatus=NOT_OWNED;
  }

//*************************************************************************************//
//**************************** GET TOKEN **********************************************//
//*************************************************************************************//

  public synchronized Token getToken()
  {
    Trace.code("TokenState.java -> synchronized Token getToken ( )");
    return token;
  }

//*************************************************************************************//
//**************************** CHANGE TOKEN *******************************************//
//*************************************************************************************//

  public synchronized void changeToken(Token newToken)
  {
    Trace.code("TokenState.java -> synchronized void changeToken ( Token newToken )");
    token = newToken;
  }

//*************************************************************************************//
//**************************** TOKEN RECEIVED *****************************************//
//*************************************************************************************//

  public synchronized void tokenReceived()
  {
    Trace.code("TokenState.java -> synchronized void tokenReceived ( )");
    assert tokenStatus != OWNED;
    tokenStatus = OWNED;
  }

//*************************************************************************************//
//**************************** SEND TOKEN *********************************************//
//*************************************************************************************//

  /**
    * Sends the token using the ChannelManager. Returns true if the sent is succesfull
    **/
  public boolean sendToken()
  {
    Trace.code("TokenState.java -> boolean sendToken ( )");
    assert channelsManager !=null ;

    synchronized(this)
    {
      assert tokenStatus == OWNED;
      tokenStatus = BEING_SENT;
    }

    boolean ret = channelsManager.sendToken(token);

    synchronized(this)
    {
      if (ret)
      {
        //it is possible that by the time the token is sent it is already received!
        if (tokenStatus==BEING_SENT)
          tokenStatus = NOT_OWNED;
      }
      else
      {
        assert tokenStatus == BEING_SENT;
        tokenStatus = OWNED;
      }
    }

    return ret;
  }

//*************************************************************************************//
//**************************** IS TOKEN OWNED OR BEING SENT ***************************//
//*************************************************************************************//

  public boolean isTokenOwnedOrBeingSent()
  {
    Trace.code("TokenState.java -> boolean isTokenOwnedOrBeingSent ( )");
    return tokenStatus!=NOT_OWNED;
  }

//*************************************************************************************//
//**************************** IS TOKEN OWNED *****************************************//
//*************************************************************************************//

  public boolean isTokenOwned()
  {
    Trace.code("TokenState.java -> boolean isTokenOwned ( )");
    return tokenStatus==OWNED;
  }

//*************************************************************************************//
//**************************** IS SAME TOKEN ******************************************//
//*************************************************************************************//

  /**
    * Returns true if the token specified matchs the one stored
    **/
  public boolean isSameToken(Token right)
  {
    Trace.code("TokenState.java -> boolean isSameToken ( Token right )");
    return token.id==right.id && token.creator==right.creator;
  }

//*************************************************************************************//
//**************************** GENERATE NEW TOKEN *************************************//
//*************************************************************************************//

  /**
    * Generates a new token
    **/
  public Token generateNewToken()
  {
    Trace.code("TokenState.java -> Token generateNewToken ( )");
    return lastGeneratedToken=new Token(lastGeneratedToken.creator, lastGeneratedToken.id+1);
  }

//*************************************************************************************//
//**************************** FREE RESOURCES *****************************************//
//*************************************************************************************//

  public void freeResources()
  {
    Trace.code("TokenState.java -> void freeResources ( )");
    channelsManager=null;
  }

//*************************************************************************************//
//**************************** DATA MEMBERS *******************************************//
//*************************************************************************************//

  ViewHandler channelsManager;	//used to send the token
  int tokenStatus;
  Token token=null;																	//current token
  Token lastGeneratedToken=null;										//last token generated by this member

  static final int OWNED 				= 0;
  static final int BEING_SENT 	= 1;
  static final int NOT_OWNED 		= 2;

}
