/**
  * File: Error.java
  * Content: Basic error facilities
  * Author: LuisM Pena
  * Date: 19th August 2000
  * Version: 0.97.00
  * Last change:
  *
  *
  **/

package sensei.util;

import sensei.util.logging.Logger;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;

/**
  * Error is a utility class which allows the code to redirect errors easily.
  * It has two main methods, error and unhandledException, to show the error conditions.
  * The output can be controlled in three ways. By default, it is sent to the System.err,
  * but this can be avoided with useStandardOutput method. It can use as well the Logger,
  * with useLogger method, and finally, it can be redirected to some Error catcher implementing
  * the ErrorHandler interface.
  **/
public class Error
{
  /**
    * Defines wether to use the standard output error to show errors. By default, this
    * is set to true
    **/
  public static void useStandardOutput(boolean use)
  {
    Error.useErr = use;
  }

  /**
    * Sets the Logger to receive the debug messages. (By default is true)
    **/
  public static void useLogger(boolean use)
  {
    Error.useLog = use;
  }

  /**
    * Sets the output to receive the debug messages. Returns the previous output installed.
    * If there is an error handler
    **/
  public static ErrorHandler setErrorHandler(ErrorHandler handler)
  {
    ErrorHandler old = Error.handler;
    Error.handler = handler;
    return old;
  }

  /**
    * raises a verified error
    **/
  public static void error(String area, String reason)
  {
    boolean handled = false;
    if (handler!=null)
      handled = handler.handleError(area, reason);
    if (!handled)
    {
      String error="Execution error ["+area+"]: " + reason;
      message(area, error);
      throw new RuntimeException(error);
    }
  }

  /**
    * Handles unexpected exceptions
    **/
  public static void unhandledException(String area, Exception ex)
  {
    boolean handled = false;
    //no reason to throw again now the exception!!!!
    if (handler!=null)
      handled = handler.handleException(area, ex);
    if (!handled)
      message(area, "UnhandledException [" + area+ "]: "+convertExceptionToString(ex));
  }

  public static String convertExceptionToString(Exception ex)
  {
    ByteArrayOutputStream byteArray = new ByteArrayOutputStream();
    ex.printStackTrace (new PrintStream(byteArray));
    return byteArray.toString();
  }

  static void message(String area, String s)
  {
    if (useErr)
      System.err.println(s);

    if (useLog)
    {
      Logger logger = Logger.getLogger(area, true);
      if (logger != null)
        logger.log(s);
    }
  }


  static ErrorHandler handler=null;
  static boolean useErr = true;
  static boolean useLog = true;
}
