/**
  * File: Parameters.java
  * Content: Handles command line parameters
  * Author: LuisM Pena
  * Date: 21st October 2000
  * Version: 0.97.00
  * Last change:
  *
  **/

package sensei.util;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

/**
  * Class to handle command line parameters. It follows a simple pattern:
  * each parameter starts always with '-', and must be followed by a value,
  * separated by a '=', for example -log=Logger.txt or -log (no value associated)
  * An special case are the definition of values, which always start by -D or -d
  * and can have the syntax -Dvariable value or -Dvariable=value
  * Parameters are then distributed in three parts:
  *  -parameters : pairs(parameter, value)
  *  -definitions: pairs(definition, value)
  *  -arguments
  * It is possible to define the parameters and definitions allowed, as well as the number
  * of arguments
  **/
public class Parameters
{
  /**
    * @param args The command line arguments
    * @param allowedParameters The parameters allowed. If null, any is allowed
    * @param allowedDefinitions The definitions allowed. If null, any is allowed
    * @param minArgs the minimum number of arguments
    * @param maxArgs the maximum number of arguments
    **/
  public Parameters(String args[], Vector allowedParameters, Vector allowedDefinitions,
                    int minArg, int maxArg) throws ParameterException
  {
    arguments = new Vector();
    parameters = new HashMap();
    definitions = new HashMap();

    validParameters = allowedParameters ;
    validDefinitions = allowedDefinitions;

    parseArguments(args);

    if (arguments.size()<minArg || arguments.size()>maxArg)
      throw new ParameterException(arguments.size());
  }

//*************************************************************************************//
//**************************** GET ARGUMENTS LENGTH ***********************************//
//*************************************************************************************//

  public int getArgumentsLength()
  {
    return arguments.size();
  }

//*************************************************************************************//
//**************************** GET ARGUMENT *******************************************//
//*************************************************************************************//

  public String getArgument(int index) throws ParameterException
  {
    if (index >= getArgumentsLength())
      throw new ParameterException(ParameterException.ARGUMENT_NOT_DEFINED, index);

    return (String) arguments.elementAt(index);
  }

//*************************************************************************************//
//**************************** GET ARGUMENTS ******************************************//
//*************************************************************************************//

  public String[] getArguments()
  {
    String ret[] = new String[getArgumentsLength()];
    arguments.toArray(ret);
    return ret;
  }

//*************************************************************************************//
//**************************** GET PARAMETERS LENGTH **********************************//
//*************************************************************************************//

  public int getParametersLength()
  {
    return parameters.size();
  }

//*************************************************************************************//
//**************************** GET PARAMETER  *****************************************//
//*************************************************************************************//

  public String getParameter(int index) throws ParameterException
  {
    if (index >= getParametersLength())
      throw new ParameterException(ParameterException.PARAMETER_NOT_DEFINED, index);

    //not very optimized!, but this shouldn't be usually called
    return (String) parameters.keySet().toArray()[index];
  }

//*************************************************************************************//
//**************************** GET PARAMETER ******************************************//
//*************************************************************************************//

  public String getParameter(String key) throws ParameterException
  {
    if (!hasParameter(key))
      throw new ParameterException(key, ParameterException.PARAMETER_NOT_DEFINED);

    return (String) parameters.get(key);
  }

//*************************************************************************************//
//**************************** GET PARAMETER ******************************************//
//*************************************************************************************//

  public String getParameter(String key, String defaultValue)
  {
    return hasParameter(key) ? (String) parameters.get(key) : defaultValue;
  }

//*************************************************************************************//
//**************************** GET PARAMETERS  ****************************************//
//*************************************************************************************//

  public Map getParameters()
  {
    return parameters;
  }

//*************************************************************************************//
//**************************** HAS PARAMETER ******************************************//
//*************************************************************************************//

  public boolean hasParameter(String key)
  {
    return parameters.containsKey(key);
  }

//*************************************************************************************//
//**************************** GET DEFINITIONS LENGTH *********************************//
//*************************************************************************************//

  public int getDefinitionsLength()
  {
    return definitions.size();
  }

//*************************************************************************************//
//**************************** GET DEFINITION *****************************************//
//*************************************************************************************//

  public String getDefinition(int index) throws ParameterException
  {
    if (index >= getDefinitionsLength())
      throw new ParameterException(ParameterException.DEFINITION_NOT_DEFINED, index);

    //not very optimized!, but this shouldn't be usually called
    return (String) definitions.keySet().toArray()[index];
  }

//*************************************************************************************//
//**************************** GET DEFINITION *****************************************//
//*************************************************************************************//

  public String getDefinition(String key) throws ParameterException
  {
    if (!hasDefinition(key))
      throw new ParameterException(key, ParameterException.DEFINITION_NOT_DEFINED);

    return (String) definitions.get(key);
  }

//*************************************************************************************//
//**************************** GET DEFINITIONS ****************************************//
//*************************************************************************************//

  public Map getDefinitions()
  {
    return definitions;
  }

//*************************************************************************************//
//**************************** SET DEFINITION *****************************************//
//*************************************************************************************//

  public void setDefinition(String key, String value)
  {
    definitions.put(key, value);
  }

//*************************************************************************************//
//**************************** HAS DEFINITION *****************************************//
//*************************************************************************************//

  public boolean hasDefinition(String key)
  {
    return definitions.containsKey(key);
  }

//*************************************************************************************//
//**************************** PARSE ARGUMENTS ****************************************//
//*************************************************************************************//

  void parseArguments(String args[]) throws ParameterException
  {
    int l = args.length;
    boolean stillParameter = true;
    for (int i=0; i < l; i++)
      if (stillParameter && args[i].charAt(0)=='-')
      {
        handleNoArgument(args[i]);
      }
      else
      {
        stillParameter = false;
        arguments.add(args[i]);
      }
  }

//*************************************************************************************//
//**************************** HANDLE NO ARGUMENT *************************************//
//*************************************************************************************//

  /**
    * Handles each parameter, throwing an exception is has a wrong format
    **/
  void handleNoArgument(String argument) throws ParameterException
  {
    int l = argument.length();

    if (l<=1)
      throw new ParameterException(argument);	//parameters is '-'

    //isDefinition: argument is a definition and not a parameter
    boolean isDefinition = argument.charAt(1)=='D';
    if (isDefinition && l==2)
      throw new ParameterException(argument);	//parameters is '-D'

    //equal: character holding the '='
    int equal = argument.indexOf('=');
    if (argument.indexOf('=',equal+1)!=-1)
      throw new ParameterException(argument);	//argument contains two or more '='

    //withValue: true if there is a character '='
    boolean withValue = equal!=-1;
    String key = argument.substring(isDefinition? 2 : 1, withValue? equal : l);
    String value = withValue? argument.substring(equal+1) : "";

    Object previous;
    if (isDefinition)
    {
      if (validDefinitions != null && !validDefinitions.contains(key))
        throw new ParameterException(key, ParameterException.NOT_ALLOWED);
      previous = definitions.put(key, value);
    }
    else
    {
      if (validParameters != null && !validParameters.contains(key))
        throw new ParameterException(key, ParameterException.NOT_ALLOWED);
      previous = parameters.put(key, value);
    }

    //cannot be defined twice any parameter or definition
    if (previous != null)
      throw new ParameterException(key, ParameterException.ALREADY_SET);
  }

//*************************************************************************************//
//**************************** DATA MEMBERS *******************************************//
//*************************************************************************************//

  Vector arguments, validParameters, validDefinitions;
  HashMap parameters, definitions;
}
