/**
  * File: FileLogStream.java
  * Content: PrintWriter that closes itself at program termination, and with some functionality
  *      			to be created as temporal file and later renamed depending on the application being
  *						logged
  * Author: LuisM Pena
  * Date: 7th November 2000
  * Version: 0.97.00
  * Last change:
  *
  **/

package sensei.util.logging;

import sensei.util.Error;
import sensei.util.ShutdownHook;
import sensei.util.ShutdownHooker;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;

class FileLogStream implements LogStream, ShutdownHooker
{

  /*******************************************************************************/
  /**************************** CONSTRUCTOR **************************************/
  /*******************************************************************************/

  /**
    * A temporal file is created, which can be later renamed using setLogName.
    * If the name is not changed, the parameter echoFileName flags when the
    * name of the file created must be shown or not (when the file is to be ended)
    **/
  public FileLogStream(boolean echoFileName)
  {
    file=null;
    writer=null;
    finalFile = null;
    nameGiven = false;
    this.echoFileName = echoFileName;
    try
    {
      file = finalFile = File.createTempFile(DEFAULT_PREFIX, DEFAULT_SUFFIX,new File("."));
      writer=new PrintWriter(new FileOutputStream(file));
      ShutdownHook.register(this);
    }
    catch(IOException ex)
    {
      close();
      System.out.println("Logging local not possible. " + ex.getMessage());
    }
  }

  /*******************************************************************************/
  /**************************** TO STRING ****************************************/
  /*******************************************************************************/

  public String toString()
  {
    if (nameGiven)
      return file.toString() + "(" + finalFile.toString() +")";
    else
      return file.toString();
  }

  /*******************************************************************************/
  /**************************** SHUTDOWN HOOKER INTERFACE ************************/
  /*******************************************************************************/

  public void shuttingDown()
  {
    close();
  }

  /*******************************************************************************/
  /**************************** LOG STREAM INTERFACE *****************************/
  /*******************************************************************************/

  public void setLogName(String name)
  {
    finalFile = new File(name+DEFAULT_SUFFIX);
    nameGiven = true;
  }

  public void println(String line)
  {
    if (writer!=null)
      writer.println(line);
  }

  public synchronized void close()
  {
    if (writer!=null)
      try
      {
        writer.close();
        writer=null;
        if (nameGiven)
        {
          finalFile.delete();
          file.renameTo(finalFile);
        }
        else if (echoFileName)
          System.out.println("Created logging file with name " + file.getName());
        file=finalFile=null;
      }
      catch(Exception ex)
      {
        Error.unhandledException(Logger.LOGGING, ex);
      }
  }

  File file;
  File finalFile;
  PrintWriter writer;
  boolean nameGiven, echoFileName;
  static final String DEFAULT_SUFFIX = ".log";
  static final String DEFAULT_PREFIX = "sensei";

}
