/**
  * File: SocketLogStream.java
  * Content: LogStream sending all the information to a specific socket
  * Author: LuisM Pena
  * Date: 7th November 2000
  * Version: 0.97.00
  * Last change:
  *
  **/

package sensei.util.logging;

import sensei.util.Error;

import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;

/**
  * A socket log stream redirects any logging to a socket.
  * Because the work to do is considered secondary, any error
  * on the socket connection is completely transparent, and just
  * an error flag is set
  **/
class SocketLogStream implements LogStream
{
  /*******************************************************************************/
  /**************************** CONSTRUCTOR **************************************/
  /*******************************************************************************/

  /**
    * The logging can only be performed on the current machine!!!
    * Therefore, the host cannot be specified.
    **/
  public SocketLogStream (int socket)
  {
    server=null;
    output=null;
    try
    {
      server = new Socket(InetAddress.getLocalHost(), socket);
      output = new DataOutputStream(server.getOutputStream());
      ok=initiateSession();
    }
    catch(Exception ex) //IOException, SecurityException
    {
      System.out.println("Logging remote on port " + socket + " not possible. " + ex.getMessage());
      ok=false;
    }
  }

  /*******************************************************************************/
  /**************************** SET LOG NAME *************************************/
  /*******************************************************************************/

  public void setLogName(String name)
  {
    if (ok)
      try
      {
        output.writeInt(LoggingSession.CHANGE_NAME);
        output.writeUTF(name);
      }
      catch(IOException ex)
      {
        Error.unhandledException(Logger.LOGGING, ex);
        ok=false;
        close();
      }
  }

  /*******************************************************************************/
  /**************************** CLOSE ********************************************/
  /*******************************************************************************/

  public synchronized void close()
  {
    if (server!=null)
    {
      if (output!=null) try{output.close();}catch(IOException ex){}
      try{server.close();}catch(IOException ex){}
      server=null;
      output=null;
    }
    ok=false;
  }

  /*******************************************************************************/
  /**************************** PRINTLN ******************************************/
  /*******************************************************************************/

  public void println(String line)
  {
    if (ok)
      try
      {
        output.writeInt(LoggingSession.PRINT);
        output.writeUTF(line);
      }
      catch(IOException ex)
      {
        Error.unhandledException(Logger.LOGGING, ex);
        ok=false;
        close();
      }
  }

  /*******************************************************************************/
  /**************************** INITIATE SESSION**********************************/
  /*******************************************************************************/

  boolean initiateSession()
  {
    ok=true;
    try
    {
      output.writeInt(LoggingSession.INIT);
      output.writeUTF(LoggingSession.PASSWORD);
    }
    catch(IOException ex)
    {
      System.out.println("Logging remote not possible. " + ex.getMessage());
      ok=false;
      close();
    }
    return ok;
  }

  Socket server;
  DataOutputStream output;
  boolean ok;
}
